"""
User-level analysis commands, typically for measuring or generating SheetViews.

Most of this file consists of commands for creating SheetViews, paired
with a template for how to construct plots from these SheetViews.

For instance, the implementation of Activity plots consists of the
update_activity() command plus the Activity PlotGroupTemplate.  The
update_activity() command reads the activity array of each Sheet and
makes a corresponding SheetView to put in the Sheet's sheet_views
dictionary, while the Activity PlotGroupTemplate specifies which
SheetViews should be plotted in which combination.  See the help for
PlotGroupTemplate for more information.

Some of the commands are ordinary Python functions, but the rest are
ParameterizedFunctions, which act like Python functions but support
Parameters with defaults, bounds, inheritance, etc.  These commands
are usually grouped together using inheritance so that they share a
set of parameters and some code, and only the bits that are specific
to that particular plot or analysis appear below.  See the
superclasses for the rest of the parameters and code.

$Id$
"""

import Image,ImageDraw
    
import copy

from numpy.oldnumeric import array, maximum
from numpy import pi, sin, cos, nonzero, round, linspace

import param
from param.parameterized import ParameterizedFunction
from param.parameterized import ParamOverrides

import topo
from topo.base.cf import Projection
from topo.base.sheet import Sheet
from topo.sheet import GeneratorSheet
from topo.base.sheetview import SheetView
from topo.misc.distribution import Distribution, DistributionStatisticFn
from topo.misc.distribution import DSF_MaxValue, DSF_BimodalPeaks
from topo.misc.distribution import DSF_WeightedAverage, DSF_VonMisesFit, DSF_BimodalVonMisesFit
from topo.pattern import GaussiansCorner, Gaussian, RawRectangle, Composite, Constant
from topo.analysis.featureresponses import ReverseCorrelation
from topo.plotting.plotgroup import create_plotgroup, plotgroups

from topo.plotting.plotgroup import UnitMeasurementCommand,ProjectionSheetMeasurementCommand
from topo.analysis.featureresponses import Feature, PatternPresenter, MeasureResponseCommand
from topo.analysis.featureresponses import SinusoidalMeasureResponseCommand, PositionMeasurementCommand, SingleInputResponseCommand
from topo.base.patterngenerator import PatternGenerator
from topo.command import pattern_present

from topo.misc.patternfn import line


# Helper function for save_plotgroup
def _equivalent_for_plotgroup_update(p1,p2):
    """
    Helper function for save_plotgroup.
    
    Comparison operator for deciding whether make_plots(update==False) is
    safe for one plotgroup if the other has already been updated.
    
    Treats plotgroups as the same if the specified list of attributes
    (if present) match in both plotgroups.
    """

    attrs_to_check = ['pre_plot_hooks','keyname','sheet','x','y','projection','input_sheet','density','coords']

    for a in attrs_to_check:
        if hasattr(p1,a) or hasattr(p2,a):
            if not (hasattr(p1,a) and hasattr(p2,a) and getattr(p1,a)== getattr(p2,a)):
                return False

    return True



class save_plotgroup(ParameterizedFunction):
    """
    Convenience command for saving a set of plots to disk.  Examples:

      save_plotgroup("Activity")
      save_plotgroup("Orientation Preference")
      save_plotgroup("Projection",projection=topo.sim['V1'].projections('Afferent'))
                                  
    Some plotgroups accept optional parameters, which can be passed
    like projection above.
    """

    equivalence_fn = param.Callable(default=_equivalent_for_plotgroup_update,doc="""
        Function to call on plotgroups p1,p2 to determine if calling pre_plot_hooks
        on one of them is sufficient to update both plots.  Should return False
        unless the commands are exact equivalents, including all relevant parameters.""")

    use_cached_results = param.Boolean(default=False,doc="""
        If True, will use the equivalence_fn to determine cases where
        the pre_plot_hooks for a plotgroup can safely be skipped, to
        avoid lengthy redundant computation.  Should usually be
        False for safety, but can be enabled for e.g. batch mode
        runs using a related batch of plots.""")

    saver_params = param.Dict(default={},doc="""
        Optional parameters to pass to the underlying PlotFileSaver object.""")


    # Class variables to cache values from previous invocations
    previous_time=[-1]
    previous_plotgroups=[]

    def __call__(self,name,**params):
        p=ParamOverrides(self,params,allow_extra_keywords=True)

        plotgroup = copy.deepcopy(plotgroups[name])
    
        # JABALERT: Why does a Projection plot need a Sheet parameter?
        # CB: It shouldn't, of course, since we know the sheet when we have
        # the projection object - it's just leftover from when we passed the
        # names instead. There should be an ALERT already about this somewhere
        # in projectionpanel.py or plotgroup.py (both need to be changed).
        if 'projection' in params:
            setattr(plotgroup,'sheet',params['projection'].dest)
    
        plotgroup._set_name(name)
        
        # Specified parameters that aren't parameters of
        # save_plotgroup() are set on the plotgroup
        for n,v in p.extra_keywords().items():
            plotgroup.set_param(n,v)

        # Reset plot cache when time changes
        if (topo.sim.time() != self.previous_time[0]):
            del self.previous_time[:]
            del self.previous_plotgroups[:]
            self.previous_time.append(topo.sim.time())
            
        # Skip update step if equivalent to prior command at this sim time
        update=True
        if p.use_cached_results:
            for g in self.previous_plotgroups:
                if p.equivalence_fn(g,plotgroup):
                    update=False
                    break

        keywords=" ".join(["%s" % (v.name if isinstance(v,param.Parameterized) else str(v)) for n,v in p.extra_keywords().items()])
        plot_description="%s%s%s" % (plotgroup.name," " if keywords else "",keywords)
        if update:
            self.previous_plotgroups.append(plotgroup)
            self.debug("%s: Running pre_plot_hooks" % plot_description)
        else:
            self.message("%s: Using cached results from pre_plot_hooks" % plot_description)

        plotgroup.make_plots(update=update)
        plotgroup.filesaver.save_to_disk(**(p.saver_params))



def decode_feature(sheet, preference_map = "OrientationPreference", axis_bounds=(0.0,1.0), cyclic=True, weighted_average=True, cropfn=lambda(x):x):
    """
    Estimate the value of a feature from the current activity pattern on a sheet.
    
    The specified preference_map should be measured before this
    function is called.
    
    If weighted_average is False, the feature value returned is the
    value of the preference_map at the maximally active location.
    
    If weighted_average is True, the feature value is estimated by
    weighting the preference_map by the current activity level, and
    averaging the result across all units in the sheet.  The
    axis_bounds specify the allowable range of the feature values in
    the preference_map.  If cyclic is true, a vector average is used;
    otherwise an arithmetic weighted average is used.
    
    For instance, if preference_map is OrientationPreference (a cyclic
    quantity), then the result will be the vector average of the
    activated orientations.  For an orientation map this value should
    be an estimate of the orientation present on the input.

    If desired, a cropfn can be supplied that will narrow the analysis
    down to a specific region of the array; this function will be
    applied to the preference_map and to the activity array before
    decoding.  Examples:

    Decode whole area:
 
       decode_feature(topo.sim["V1"])
    
    Decode left half only:

       r,c = topo.sim["V1"].activity.shape
       lefthalf  = lambda(x): x[:,0:c/2]
       righthalf = lambda(x): x[:,c/2:]

       decode_feature(topo.sim["V1"], cropfn=lefthalf)
    
    """

    d = Distribution(axis_bounds, cyclic)
    
    if not (preference_map in sheet.sheet_views):
        topo.sim.warning(preference_map + " should be measured before calling decode_feature.")
    else:
        v = sheet.sheet_views[preference_map]
        d.add(dict(zip(cropfn(v.view()[0]).ravel(), 
                       cropfn(sheet.activity).ravel())))
    
    res = DSF_WeightedAverage()(d) if weighted_average else DSF_MaxValue()(d)
    return res['']['preference']
    


def update_activity():
    """
    Make a map of neural activity available for each sheet, for use in template-based plots.
    
    This command simply asks each sheet for a copy of its activity
    matrix, and then makes it available for plotting.  Of course, for
    some sheets providing this information may be non-trivial, e.g. if
    they need to average over recent spiking activity.
    """
    for sheet in topo.sim.objects(Sheet).values():
        activity_copy = array(sheet.activity)
        new_view = SheetView((activity_copy,sheet.bounds),
                              sheet.name,sheet.precedence,topo.sim.time(),sheet.row_precedence)
        sheet.sheet_views['Activity']=new_view


pg = create_plotgroup(name='Activity',category='Basic',
             doc='Plot the activity for all Sheets.', auto_refresh=True,
             pre_plot_hooks=[update_activity], plot_immediately=True)
pg.add_plot('Activity',[('Strength','Activity')])


def update_rgb_activities():
    """
    Make available Red, Green, and Blue activity matrices for all appropriate sheets.
    """
    for sheet in topo.sim.objects(Sheet).values():
        for c in ['Red','Green','Blue']:
            # should this ensure all of r,g,b are present? 
            if hasattr(sheet,'activity_%s'%c.lower()): 
                activity_copy = getattr(sheet,'activity_%s'%c.lower()).copy()
                new_view = SheetView((activity_copy,sheet.bounds),
                                     sheet.name,sheet.precedence,topo.sim.time(),sheet.row_precedence)
                sheet.sheet_views['%sActivity'%c]=new_view
            

pg = create_plotgroup(name='RGB',category='Other',
             doc='Combine and plot the red, green, and blue activity for all appropriate Sheets.', auto_refresh=True,
             pre_plot_hooks=[update_rgb_activities], plot_immediately=True)
pg.add_plot('RGB',[('Red','RedActivity'),('Green','GreenActivity'),('Blue','BlueActivity')])



class update_connectionfields(UnitMeasurementCommand):
    """A callable Parameterized command for measuring or plotting a unit from a Projection."""

    # Force plotting of all CFs, not just one Projection
    projection = param.ObjectSelector(default=None,constant=True)


pg= create_plotgroup(name='Connection Fields',category="Basic",
                     doc='Plot the weight strength in each ConnectionField of a specific unit of a Sheet.',
                     pre_plot_hooks=[update_connectionfields],
                     plot_immediately=True, normalize='Individually', situate=True)
pg.add_plot('Connection Fields',[('Strength','Weights')])



class update_projection(UnitMeasurementCommand):
    """A callable Parameterized command for measuring or plotting units from a Projection."""


pg= create_plotgroup(name='Projection',category="Basic",
           doc='Plot the weights of an array of ConnectionFields in a Projection.',
           pre_plot_hooks=[update_projection],
           plot_immediately=False, normalize='Individually',sheet_coords=True)
pg.add_plot('Projection',[('Strength','Weights')])



class update_projectionactivity(ProjectionSheetMeasurementCommand):
    """
    Add SheetViews for all of the Projections of the ProjectionSheet
    specified by the sheet parameter, for use in template-based plots.
    """
    
    def __call__(self,**params):
        p=ParamOverrides(self,params)
        self.params('sheet').compute_default()        
        s = p.sheet
        if s is not None:
            for conn in s.in_connections:
                if not isinstance(conn,Projection):
                    topo.sim.debug("Skipping non-Projection "+conn.name)
                else:
                    v = conn.get_projection_view(topo.sim.time())
######################################################################
## CEBALERT: when a TemplatePlot tp is created from a ProjectionView
## pv, tp.plot_src_name comes ultimately from pv.projection.src.name,
## but tp.plot_bounding_box comes from pv.projection.dest.bounds. So,
## we get a plot with mismatched src_name and bounds. Fixing that here
## is not the correct solution, but it allows the Projection Activity
## GUI window to work.
                    v.src_name = v.projection.dest.name
## JABALERT: Similarly, probably not the right thing to do, but makes
## the name of the src available for the plot to be labelled.
                    v.proj_src_name = v.projection.src.name
######################################################################
                    key = ('ProjectionActivity',v.projection.dest.name,v.projection.name)
                    v.projection.dest.sheet_views[key] = v


pg =  create_plotgroup(name='Projection Activity',category="Basic",
             doc='Plot the activity in each Projection that connects to a Sheet.',
             pre_plot_hooks=[update_projectionactivity.instance()],
             plot_immediately=True, normalize='Individually',auto_refresh=True)
pg.add_plot('Projection Activity',[('Strength','ProjectionActivity')])


class measure_rfs(SingleInputResponseCommand):
    """
    Map receptive fields by reverse correlation.

    Presents a large collection of input patterns, typically pixel
    by pixel on and off, keeping track of which units in the specified
    input_sheet were active when each unit in other Sheets in the
    simulation was active.  This data can then be used to plot
    receptive fields for each unit.  Note that the results are true
    receptive fields, not the connection fields usually presented in
    lieu of receptive fields, because they take all circuitry in
    between the input and the target unit into account.

    Note also that it is crucial to set the scale parameter properly when
    using units with a hard activation threshold (as opposed to a
    smooth sigmoid), because the input pattern used here may not be a
    very effective way to drive the unit to activate.  The value
    should be set high enough that the target units activate at least
    some of the time there is a pattern on the input.
    """  
    static_parameters = param.List(default=["offset","size"])
    
    sampling = param.Number(default=1.0,bounds=(1.0,None),doc="""
    	The sampling value determines the size of the input pattern and
    	the number of presentations required to fully sample the input sheet.
    	The higher the value the coarser the receptive field measurement 
    	will be.""")
    
    area_ratio = param.Number(default=1.0,bounds=(0.0,1.0),doc="""
    	Ratio defining the area in the input sheet that is sampled during
    	the reverse correlation procedure. Reducing this value below 1.0
    	will invalidate the RFs of all neurons outside the specified area.""")

    __abstract = True
    
    def __call__(self,**params):
        p=ParamOverrides(self,params)
        self.params('input_sheet').compute_default()
        x=ReverseCorrelation(self._feature_list(p),input_sheet=p.input_sheet)
        static_params = dict([(s,p[s]) for s in p.static_parameters])
    
        if p.duration is not None:
            p.pattern_presenter.duration=p.duration
        if p.apply_output_fns is not None:
            p.pattern_presenter.apply_output_fns=p.apply_output_fns
        x.collect_feature_responses(p.pattern_presenter,static_params,p.display,self._feature_list(p))
          
    def _feature_list(self,p):
    
        left, bottom, right, top = p.input_sheet.nominal_bounds.lbrt()
        left *= p.area_ratio; bottom *= p.area_ratio; right *= p.area_ratio; top *= p.area_ratio
        sheet_density = float(p.input_sheet.nominal_density)
        
        # Cannot assume square sheet.
        vertical_divisions = ((sheet_density * (top - bottom)) - 1) / p.sampling
        horizontal_divisions = ((sheet_density * (right - left)) - 1) / p.sampling

        unit_size = 1.0 / sheet_density * p.sampling
        half_unit_size = unit_size / 2.0 # saves repeated calculation.
        p['size'] = unit_size
        
        # Set the x and y max values down by half a unit so patterns are presented in the centre of each unit.
        y_range = (top - half_unit_size, bottom)
        x_range = (right - half_unit_size, left)

        return [Feature(name="x", range=x_range, step=float(x_range[1]-x_range[0])/horizontal_divisions), 
                Feature(name="y", range=y_range, step=float(y_range[1]-y_range[0])/vertical_divisions),
                Feature(name="scale", range=(-p.scale, p.scale), step=p.scale*2)]
        
pg = create_plotgroup(name='RF Projection',category='Other',
    doc='Measure receptive fields.',
    pre_plot_hooks=[measure_rfs.instance(display=True, 
    pattern_presenter=PatternPresenter(RawRectangle(size=0.01,aspect_ratio=1.0)))],
    normalize='Individually')

pg.add_plot('RFs',[('Strength','RFs')])
 
               
# Helper function for measuring direction maps
def compute_orientation_from_direction(current_values):
    """ 
    Return the orientation corresponding to the given direction. 

    Wraps the value to be in the range [0,pi), and rounds it slightly
    so that wrapped values are precisely the same (to avoid biases
    caused by vector averaging with keep_peak=True).

    Note that in very rare cases (1 in 10^-13?), rounding could lead
    to different values for a wrapped quantity, and thus give a
    heavily biased orientation map.  In that case, just choose a
    different number of directions to test, to avoid that floating
    point boundary.
    """
    return round(((dict(current_values)['direction'])+(pi/2)) % pi,13)



class measure_sine_pref(SinusoidalMeasureResponseCommand):
    """
    Measure preferences for sine gratings in various combinations.
    Can measure orientation, spatial frequency, spatial phase,
    ocular dominance, horizontal phase disparity, color hue, motion
    direction, and speed of motion.

    In practice, this command is useful for any subset of the possible
    combinations, but if all combinations are included, the number of
    input patterns quickly grows quite large, much larger than the
    typical number of patterns required for an entire simulation.
    Thus typically this command will be used for the subset of
    dimensions that need to be evaluated together, while simpler
    special-purpose routines are provided below for other dimensions
    (such as hue and disparity).
    """

    num_ocularity = param.Integer(default=1,bounds=(1,None),softbounds=(1,3),doc="""
        Number of ocularity values to test; set to 1 to disable or 2 to enable.""")

    num_disparity = param.Integer(default=1,bounds=(1,None),softbounds=(1,48),doc="""
        Number of disparity values to test; set to 1 to disable or e.g. 12 to enable.""")

    num_hue = param.Integer(default=1,bounds=(1,None),softbounds=(1,48),doc="""
        Number of hues to test; set to 1 to disable or e.g. 8 to enable.""")

    num_direction = param.Integer(default=0,bounds=(0,None),softbounds=(0,48),doc="""
        Number of directions to test.  If nonzero, overrides num_orientation, 
        because the orientation is calculated to be perpendicular to the direction.""")

    num_speeds = param.Integer(default=4,bounds=(0,None),softbounds=(0,10),doc="""
        Number of speeds to test (where zero means only static patterns).
        Ignored when num_direction=0.""")

    max_speed = param.Number(default=2.0/24.0,bounds=(0,None),doc="""
        The maximum speed to measure (with zero always the minimum).""")

    subplot = param.String("Orientation")

    
    def _feature_list(self,p):
        # Always varies frequency and phase; everything else depends on parameters.

        features = \
            [Feature(name="frequency",values=p.frequencies,
                    preference_fn=DSF_WeightedAverage())]

        if p.num_direction==0: features += \
            [Feature(name="orientation",range=(0.0,pi),step=pi/p.num_orientation,
                            cyclic=True, preference_fn=self.preference_fn)]

        features += \
            [Feature(name="phase",range=(0.0,2*pi),step=2*pi/p.num_phase,cyclic=True,
                     preference_fn=DSF_WeightedAverage( value_scale=(0., 1/2.0/pi) ))]

        if p.num_ocularity>1: features += \
            [Feature(name="ocular",range=(0.0,1.0),step=1.0/p.num_ocularity)]

        if p.num_disparity>1: features += \
            [Feature(name="phasedisparity",range=(0.0,2*pi),step=2*pi/p.num_disparity,cyclic=True)]

        if p.num_hue>1: features += \
            [Feature(name="hue",range=(0.0,1.0),step=1.0/p.num_hue,cyclic=True)]
            
        if p.num_direction>0 and p.num_speeds==0: features += \
            [Feature(name="speed",values=[0],cyclic=False)]
        
        if p.num_direction>0 and p.num_speeds>0: features += \
            [Feature(name="speed",range=(0.0,p.max_speed),step=float(p.max_speed)/p.num_speeds,cyclic=False)]

        if p.num_direction>0:
            # Compute orientation from direction
            dr = Feature(name="direction",range=(0.0,2*pi),step=2*pi/p.num_direction,cyclic=True)
            or_values = list(set([compute_orientation_from_direction([("direction",v)]) for v in dr.values]))
            features += [dr, \
                 Feature(name="orientation",range=(0.0,pi),values=or_values,cyclic=True,
                         compute_fn=compute_orientation_from_direction,preference_fn=self.preference_fn)]

        return features


# Here as the simplest possible example; could be moved elsewhere.
class measure_or_pref(SinusoidalMeasureResponseCommand):
    """Measure an orientation preference map by collating the response to patterns."""

    subplot = param.String("Orientation")

    preference_fn = param.ClassSelector( DistributionStatisticFn, 
        default=DSF_WeightedAverage(value_scale=(0.0,1.0/pi)), doc=""" 
        Function that will be used to analyze the distributions of unit
        responses. Sets value_scale to normalize orientation preference
        values.""" )
    
    def _feature_list(self,p):

        return [Feature(name="frequency",values=p.frequencies),
                Feature(name="orientation",range=(0.0,pi),step=pi/p.num_orientation,
                        preference_fn=self.preference_fn,cyclic=True),
                Feature(name="phase",range=(0.0,2*pi),step=2*pi/p.num_phase,cyclic=True)]


pg= create_plotgroup(name='Orientation Preference',category="Preference Maps",
             doc='Measure preference for sine grating orientation.',
             pre_plot_hooks=[measure_sine_pref.instance(
                 preference_fn=DSF_WeightedAverage( value_scale=(0., 1./pi) ))] )
pg.add_plot('Orientation Preference',[('Hue','OrientationPreference')])
pg.add_plot('Orientation Preference&Selectivity',
            [('Hue','OrientationPreference'), ('Confidence','OrientationSelectivity')])
pg.add_plot('Orientation Selectivity',[('Strength','OrientationSelectivity')])
pg.add_plot('Phase Preference',[('Hue','PhasePreference')])
pg.add_plot('Phase Selectivity',[('Strength','PhaseSelectivity')])
pg.add_static_image('Color Key','command/or_key_white_vert_small.png')


pg= create_plotgroup(name='vonMises Orientation Preference',category="Preference Maps",
             doc='Measure preference for sine grating orientation using von Mises fit.',
             pre_plot_hooks=[measure_sine_pref.instance(
                 preference_fn=DSF_VonMisesFit( value_scale=(0., 1./pi) ),
                 num_orientation=16)])
pg.add_plot('Orientation Preference',[('Hue','OrientationPreference')])
pg.add_plot('Orientation Preference&Selectivity',
            [('Hue','OrientationPreference'), ('Confidence','OrientationSelectivity')])
pg.add_plot('Orientation Selectivity',[('Strength','OrientationSelectivity')])
pg.add_plot('Phase Preference',[('Hue','PhasePreference')])
pg.add_plot('Phase Selectivity',[('Strength','PhaseSelectivity')])
pg.add_static_image('Color Key','command/or_key_white_vert_small.png')


pg= create_plotgroup(name='Bimodal Orientation Preference', category="Preference Maps",
             doc='Measure preference for sine grating orientation using bimodal von Mises fit.',
             pre_plot_hooks=[measure_sine_pref.instance(
                 preference_fn=DSF_BimodalVonMisesFit( value_scale=(0., 1./pi) ),
                 num_orientation=16)])
pg.add_plot('Orientation Preference',[('Hue','OrientationPreference')])
pg.add_plot('Orientation Preference&Selectivity',
            [('Hue','OrientationPreference'), ('Confidence','OrientationSelectivity')])
pg.add_plot('Orientation Selectivity',[('Strength','OrientationSelectivity')])
pg.add_plot('Second Orientation Preference', [('Hue','OrientationMode2Preference')])
pg.add_plot('Second Orientation Preference&Selectivity',
            [('Hue','OrientationMode2Preference'), ('Confidence','OrientationMode2Selectivity')])
pg.add_plot('Second Orientation Selectivity', [('Strength','OrientationMode2Selectivity')])
pg.add_static_image('Color Key', 'command/or_key_white_vert_small.png')

        
pg = create_plotgroup(name='Two Orientation Preferences',category='Preference Maps',
    doc='Display the two most preferred orientations for each units, using bimodal von Mises fit.',
    pre_plot_hooks=[measure_sine_pref.instance(
         preference_fn=DSF_BimodalVonMisesFit(), num_orientation=16)])
pg.add_plot( 'Two Orientation Preferences', [
		( 'Or1',	'OrientationPreference' ),
		( 'Sel1',	'OrientationSelectivity' ),
		( 'Or2',	'OrientationMode2Preference' ),
		( 'Sel2',	'OrientationMode2Selectivity' )
])
pg.add_static_image('Color Key','command/two_or_key_vert.png')


pg= create_plotgroup(name='Spatial Frequency Preference',category="Preference Maps",
             doc='Measure preference for sine grating orientation and frequency.',
             pre_plot_hooks=[measure_sine_pref.instance(
                 preference_fn=DSF_WeightedAverage( value_scale=(0., 1./pi) ))] )
pg.add_plot('Spatial Frequency Preference',[('Strength','FrequencyPreference')])
pg.add_plot('Spatial Frequency Selectivity',[('Strength','FrequencySelectivity')])
# Just calls measure_sine_pref to plot different maps.


class measure_od_pref(SinusoidalMeasureResponseCommand):
    """Measure an ocular dominance preference map by collating the response to patterns."""

    def _feature_list(self,p):
        return [Feature(name="frequency",values=p.frequencies),
                Feature(name="orientation",range=(0.0,pi),step=pi/p.num_orientation,cyclic=True),
                Feature(name="phase",range=(0.0,2*pi),step=2*pi/p.num_phase,cyclic=True),
                Feature(name="ocular",range=(0.0,1.0),values=[0.0,1.0])]

pg= create_plotgroup(name='Ocular Preference',category="Preference Maps",
             doc='Measure preference for sine gratings between two eyes.',
             pre_plot_hooks=[measure_sine_pref.instance()])
pg.add_plot('Ocular Preference',[('Strength','OcularPreference')])
pg.add_plot('Ocular Selectivity',[('Strength','OcularSelectivity')])




class measure_phasedisparity(SinusoidalMeasureResponseCommand):
    """Measure a phase disparity preference map by collating the response to patterns."""

    num_disparity = param.Integer(default=12,bounds=(1,None),softbounds=(1,48),
                                  doc="Number of disparity values to test.")

    orientation = param.Number(default=pi/2,softbounds=(0.0,2*pi),doc="""
        Orientation of the test pattern; typically vertical to measure
        horizontal disparity.""")
    
    static_parameters = param.List(default=["orientation","scale","offset"])

    def _feature_list(self,p):
        return [Feature(name="frequency",values=p.frequencies),
                Feature(name="phase",range=(0.0,2*pi),step=2*pi/p.num_phase,cyclic=True),
                Feature(name="phasedisparity",range=(0.0,2*pi),step=2*pi/p.num_disparity,cyclic=True)]


pg= create_plotgroup(name='PhaseDisparity Preference',category="Preference Maps",doc="""
    Measure preference for sine gratings at a specific orentation differing in phase
    between two input sheets.""",
             pre_plot_hooks=[measure_phasedisparity.instance()],normalize='Individually')
pg.add_plot('PhaseDisparity Preference',[('Hue','PhasedisparityPreference')])
pg.add_plot('PhaseDisparity Preference&Selectivity',
            [('Hue','PhasedisparityPreference'), ('Confidence','PhasedisparitySelectivity')])
pg.add_plot('PhaseDisparity Selectivity',[('Strength','PhasedisparitySelectivity')])
pg.add_static_image('Color Key','command/disp_key_white_vert_small.png')



class measure_dr_pref(SinusoidalMeasureResponseCommand):
    """Measure a direction preference map by collating the response to patterns."""

    num_phase = param.Integer(default=12)
    
    num_direction = param.Integer(default=6,bounds=(1,None),softbounds=(1,48),
                                  doc="Number of directions to test.")

    num_speeds = param.Integer(default=4,bounds=(0,None),softbounds=(0,10),doc="""
        Number of speeds to test (where zero means only static patterns).""")

    max_speed = param.Number(default=2.0/24.0,bounds=(0,None),doc="""
        The maximum speed to measure (with zero always the minimum).""")

    subplot = param.String("Direction")

    def _feature_list(self,p):
        # orientation is computed from direction
        dr = Feature(name="direction",range=(0.0,2*pi),step=2*pi/p.num_direction,cyclic=True)
        or_values = list(set([compute_orientation_from_direction([("direction",v)]) for v in dr.values]))

        return [Feature(name="speed",values=[0],cyclic=False) if p.num_speeds is 0 else
                Feature(name="speed",range=(0.0,p.max_speed),step=float(p.max_speed)/p.num_speeds,cyclic=False),
                Feature(name="frequency",values=p.frequencies),
                Feature(name="direction",range=(0.0,2*pi),step=2*pi/p.num_direction,cyclic=True),
                Feature(name="phase",range=(0.0,2*pi),step=2*pi/p.num_phase,cyclic=True),
                Feature(name="orientation",range=(0.0,pi),values=or_values,cyclic=True,
                        compute_fn=compute_orientation_from_direction)]


pg= create_plotgroup(name='Direction Preference',category="Preference Maps",
             doc='Measure preference for sine grating movement direction.',
             pre_plot_hooks=[measure_dr_pref.instance()])
pg.add_plot('Direction Preference',[('Hue','DirectionPreference')])
pg.add_plot('Direction Preference&Selectivity',[('Hue','DirectionPreference'),
                                                ('Confidence','DirectionSelectivity')])
pg.add_plot('Direction Selectivity',[('Strength','DirectionSelectivity')])
pg.add_plot('Speed Preference',[('Strength','SpeedPreference')])
pg.add_plot('Speed Selectivity',[('Strength','SpeedSelectivity')])
pg.add_static_image('Color Key','command/dr_key_white_vert_small.png')



class measure_hue_pref(SinusoidalMeasureResponseCommand):
    """Measure a hue preference map by collating the response to patterns."""

    num_phase = param.Integer(default=12)
    
    num_hue = param.Integer(default=8,bounds=(1,None),softbounds=(1,48),
                            doc="Number of hues to test.")

    subplot = param.String("Hue")

    # For backwards compatibility; not sure why it needs to differ from the default
    static_parameters = param.List(default=[])

    def _feature_list(self,p):
        return [Feature(name="frequency",values=p.frequencies),
                Feature(name="orientation",range=(0,pi),step=pi/p.num_orientation,cyclic=True),
                Feature(name="hue",range=(0.0,1.0),step=1.0/p.num_hue,cyclic=True),
                Feature(name="phase",range=(0.0,2*pi),step=2*pi/p.num_phase,cyclic=True)]


pg= create_plotgroup(name='Hue Preference',category="Preference Maps",
             doc='Measure preference for colors.',
             pre_plot_hooks=[measure_hue_pref.instance()],normalize='Individually')
pg.add_plot('Hue Preference',[('Hue','HuePreference')])
pg.add_plot('Hue Preference&Selectivity',[('Hue','HuePreference'), ('Confidence','HueSelectivity')])
pg.add_plot('Hue Selectivity',[('Strength','HueSelectivity')])



gaussian_corner = Composite(
    operator = maximum, generators = [
        Gaussian(size = 0.06,orientation=0,aspect_ratio=7,x=0.3),
        Gaussian(size = 0.06,orientation=pi/2,aspect_ratio=7,y=0.3)])


class measure_second_or_pref(SinusoidalMeasureResponseCommand):
    """Measure the secondary  orientation preference maps."""

    num_orientation	= param.Integer( default=16, bounds=(1,None), softbounds=(1,64),
                                    doc="Number of orientations to test.")
    true_peak 	 	= param.Boolean( default=True, doc="""If set the second
	    orientation response is computed on the true second mode of the
	    orientation distribution, otherwise is just the second maximum response""" ) 

    subplot		= param.String("Second Orientation")
    
    def _feature_list(self, p):
    	fs	= [ Feature(name="frequency", values=p.frequencies) ]
    	if p.true_peak:
	    fs.append(
		Feature(name="orientation", range=(0.0, pi), step=pi/p.num_orientation,
                    cyclic=True, preference_fn=DSF_BimodalPeaks() ) )
	else:
	    fs.append(
		Feature(name="orientation", range=(0.0, pi), step=pi/p.num_orientation,
                    cyclic=True, preference_fn=DSF_BimodalPeaks() ) )
	fs.append( Feature(name="phase", range=(0.0, 2*pi), step=2*pi/p.num_phase, cyclic=True) )

	return fs


pg= create_plotgroup(name='Second Orientation Preference', category="Preference Maps",
             doc='Measure the second preference for sine grating orientation.',
             pre_plot_hooks=[measure_second_or_pref.instance( true_peak=False )])
pg.add_plot('Second Orientation Preference', [('Hue','OrientationMode2Preference')])
pg.add_plot('Second Orientation Preference&Selectivity',
            [('Hue','OrientationMode2Preference'), ('Confidence','OrientationMode2Selectivity')])
pg.add_plot('Second Orientation Selectivity', [('Strength','OrientationMode2Selectivity')])
pg.add_static_image('Color Key', 'command/or_key_white_vert_small.png')


pg= create_plotgroup(name='Second Peak Orientation Preference',category="Preference Maps",
             doc='Measure the second peak preference for sine grating orientation.',
             pre_plot_hooks=[measure_second_or_pref.instance( true_peak=True )])
pg.add_plot('Second Peak Orientation Preference', [('Hue','OrientationMode2Preference')])
pg.add_plot('Second Peak Orientation Preference&Selectivity',
            [('Hue','OrientationMode2Preference'), ('Confidence','OrientationMode2Selectivity')])
pg.add_plot('Second Peak Orientation Selectivity', [('Strength','OrientationMode2Selectivity')])
pg.add_static_image('Color Key','command/or_key_white_vert_small.png')

        
pg = create_plotgroup(name='Two Peaks Orientation Preferences',category='Preference Maps',
    doc="""Display the two most preferred orientations for all units with a
    multimodal orientation preference distribution.""",
    pre_plot_hooks=[
    		measure_second_or_pref.instance(num_orientation=16, true_peak=True)
])
pg.add_plot( 'Two Peaks Orientation Preferences', [
		( 'Or1',	'OrientationPreference' ),
		( 'Sel1',	'OrientationSelectivity' ),
		( 'Or2',	'OrientationMode2Preference' ),
		( 'Sel2',	'OrientationMode2Selectivity' )
])
pg.add_static_image('Color Key','command/two_or_key_vert.png')

class measure_corner_or_pref(PositionMeasurementCommand):
    """Measure a corner preference map by collating the response to patterns."""
    
    scale = param.Number(default=1.0)

    divisions=param.Integer(default=10)

    pattern_presenter = param.Callable(PatternPresenter(gaussian_corner,apply_output_fns=False,duration=1.0))

    x_range=param.NumericTuple((-1.2,1.2))

    y_range=param.NumericTuple((-1.2,1.2))

    num_orientation = param.Integer(default=4,bounds=(1,None),softbounds=(1,24),
                                    doc="Number of orientations to test.")

    # JABALERT: Presumably this should be omitted, so that size is included?
    static_parameters = param.List(default=["scale","offset"])

    def _feature_list(self,p):
        width =1.0*p.x_range[1]-p.x_range[0]
        height=1.0*p.y_range[1]-p.y_range[0]
        return [
            Feature(name="x",range=p.x_range,step=width/p.divisions,preference_fn=self.preference_fn),
            Feature(name="y",range=p.y_range,step=height/p.divisions,preference_fn=self.preference_fn),
            Feature(name="orientation",range=(0,2*pi),step=2*pi/p.num_orientation,cyclic=True,
                preference_fn=DSF_WeightedAverage(value_scale=(0., 1/2.0/pi))
            )]


pg= create_plotgroup(name='Corner OR Preference',category="Preference Maps",
             doc='Measure orientation preference for corner shape (or other complex stimuli that cannot be represented as fullfield patterns).',
             pre_plot_hooks=[measure_corner_or_pref.instance(
                 preference_fn=DSF_WeightedAverage())],
             normalize='Individually')
pg.add_plot('Corner Orientation Preference',[('Hue','OrientationPreference')])
pg.add_plot('Corner Orientation Preference&Selectivity',[('Hue','OrientationPreference'),
                                                   ('Confidence','OrientationSelectivity')])
pg.add_plot('Corner Orientation Selectivity',[('Strength','OrientationSelectivity')])


class measure_corner_angle_pref(PositionMeasurementCommand):
    """Generate the preference map for angle shapes, by collating the response to patterns."""
    
    scale = param.Number(default=1.0)

    size = param.Number(default=0.2)

    positions = param.Integer(default=6)

    x_range = param.NumericTuple((-1.0, 1.0))

    y_range = param.NumericTuple((-1.0, 1.0))

    num_or = param.Integer(default=4,bounds=(1,None),softbounds=(1,24),doc=
        "Number of orientations to test.")
    
    angle_0 = param.Number(default=0.25*pi,bounds=(0.0,pi),softbounds=(0.0,0.5*pi),doc=
        "First angle to test.")

    angle_1 = param.Number(default=0.75*pi,bounds=(0.0,pi),softbounds=(0.5*pi,pi),doc=
        "Last angle to test.")

    num_angle=param.Integer(default=4,bounds=(1,None),softbounds=(1,12),doc=
        "Number of angles to test.")

    key_img_fname=param.Filename(default='command/key_angles.png',doc=
        "Name of the file with the image used to code angles with hues.")

    pattern_presenter=PatternPresenter(GaussiansCorner(aspect_ratio=4.0,cross=0.85),apply_output_fns=False,duration=1.0) 

    static_parameters = param.List( default=[ "size", "scale", "offset" ] )

    

    def _feature_list( self, p ):
    	"""Return the list of features to vary, generate hue code static image"""
        x_step	= ( p.x_range[1]-p.x_range[0] ) / float( p.positions - 1 )
        y_step	= ( p.y_range[1]-p.y_range[0] ) / float( p.positions - 1 )
	o_step	= 2.0*pi / p.num_or
	if p.angle_0 < p.angle_1:
            angle_0 = p.angle_0
            angle_1 = p.angle_1
	else:
            angle_0 = p.angle_1
            angle_1 = p.angle_0
	a_range	= ( angle_0, angle_1 )
	a_step	= ( angle_1 - angle_0 ) / float( p.num_angle - 1 )
	self._make_key_image( p )
        return [
            Feature( name="x",           range=p.x_range, step=x_step ),
            Feature( name="y",           range=p.y_range, step=y_step ),
            Feature( name="orientation", range=(0, 2*pi), step=o_step, cyclic=True ),
            Feature( name="angle",       range=a_range,   step=a_step,
                    preference_fn=DSF_WeightedAverage( value_scale=( -angle_0, 1./(angle_1-angle_0) ) ) )
	]


    def _make_key_image( self, p ):
    	"""Generate the image with keys to hues used to code angles
	   the image is saved on-the-fly, in order to fit the current
	   choice of angle range
	"""
	width	= 60
	height	= 300
	border	= 6
	n_a	= 7
	angle_0	= p.angle_0
	angle_1	= p.angle_1
	a_step	= 0.5 * ( angle_1 - angle_0 ) / float( n_a )
	x_0	= border
	x_1	= ( width - border ) / 2
	x_a	= x_1 + 2 * border
	y_use	= height - 2 * border
	y_step	= y_use / float( n_a )
	y_d	= int( float( 0.5 * y_step ) )
	y_0	= border + y_d
	l	= 15

	hues	= [ "hsl(%2d,100%%,50%%)" % h		for h in range( 0, 360, 360 / n_a ) ]
	angles	= [ 0.5*angle_0 + a_step * a		for a in range( n_a ) ]
	y_pos	= [ int( round( y_0 + y * y_step ) )	for y in range( n_a ) ]
	deltas	= [ ( int( round( l * cos( a ) ) ), int( round( l * sin( a ) ) ) )
							for a in angles ]
	lb_img	= Image.new( "RGB", ( width, height ), "white" )
	dr_img	= ImageDraw.Draw( lb_img )

	for h, y, d	in zip( hues, y_pos, deltas ):
		dr_img.rectangle( [ ( x_0, y - y_d ), ( x_1, y + y_d ) ], fill = h )
		dr_img.line( [ ( x_a, y ), ( x_a + d[ 0 ], y + d[ 1 ] ) ], fill = "black" )
		dr_img.line( [ ( x_a, y ), ( x_a + d[ 0 ], y - d[ 1 ] ) ], fill = "black" )
	
	lb_img.save( p.key_img_fname )

	#return( p.key_img_fname.default )
		

pg= create_plotgroup(name='Corner Angle Preference',category="Preference Maps",
             doc='Measure preference for angles in corner shapes',
             normalize='Individually')
pg.pre_plot_hooks=[ measure_corner_angle_pref.instance() ]
pg.add_plot('Corner Angle Preference',[('Hue','AnglePreference')])
pg.add_plot('Corner Angle Preference&Selectivity',[('Hue','AnglePreference'),
                                                   ('Confidence','AngleSelectivity')])
pg.add_plot('Corner Angle Selectivity',[('Strength','AngleSelectivity')])
pg.add_plot('Corner Orientation Preference',[('Hue','OrientationPreference')])
pg.add_plot('Corner Orientation Preference&Selectivity',[('Hue','OrientationPreference'),
                                                   ('Confidence','OrientationSelectivity')])
pg.add_plot('Corner Orientation Selectivity',[('Strength','OrientationSelectivity')])
pg.add_static_image( 'Hue Code', measure_corner_angle_pref.instance().key_img_fname )



class PatternPresenter2(param.Parameterized):

    apply_output_fns = param.Boolean(default = True, 
        doc = """When presenting a pattern, whether to apply each 
        sheet's output function.  If False, for many networks the 
        response will be linear, which requires fewer test patterns to 
        measure a meaningful response, but it may not correspond to the 
        actual preferences of each neuron under other conditions. If 
        True, callers will need to ensure that the input patterns are 
        in a suitable range to drive the neurons to generate meaningful
        output, because e.g. a threshold-based output function might
        result in no activity for inputs that are too weak..""")

    duration = param.Number(default = 1.0,
        doc = """Amount of simulation time for which to present each 
        test pattern. By convention, most Topographica example files 
        are 
        designed to have a suitable activity pattern computed by the
        default time, but the duration will need to be changed for
        other models that do not follow that convention or if a
        linear response is desired.""")

    # CEBALERT: generator_sheets=[] is probably a surprising way of
    # actually getting all the generator sheets.
    generator_sheets = param.List(default = [], 
        doc="""The set of GeneratorSheets onto which patterns will be 
        drawn. By default (i.e. for an empty list), all GeneratorSheets 
        in the simulation will be used.""")
        
    pattern_generator = param.Parameter(default=Constant(),
        doc = """The PatternGenerator that will be drawn on the generator 
        sheets (the parameters of the pattern_generator are specified 
        during calls).""")


    def __call__(self, features_values, param_dict):
        
        for param, value in param_dict.iteritems():
            setattr(self.pattern_generator, param, value)                               
                              
        for feature, value in features_values.iteritems():
            setattr(self.pattern_generator, feature, value)                               
        
        all_input_sheet_names = topo.sim.objects(GeneratorSheet).keys()

        if len(self.generator_sheets) > 0:
            input_sheet_names = [sheet.name for sheet in self.generator_sheets]
        else:
            input_sheet_names = all_input_sheet_names
        
        # Copy the given generator once for every GeneratorSheet
        inputs = dict.fromkeys(input_sheet_names)
        for key in inputs.keys():
            inputs[key] = copy.deepcopy(self.pattern_generator)
            
        self._custom_presenter(inputs, input_sheet_names)
        
        # blank patterns for unused generator sheets
        for sheet_name in set(all_input_sheet_names).difference(set(input_sheet_names)):
            inputs[sheet_name] = Constant(scale=0)
                
        pattern_present(inputs, self.duration, plastic=False, apply_output_fns=self.apply_output_fns)                
                        

    def _custom_presenter(self, inputs, input_sheet_names):
        """This method provides a minimum overload for performing custom actions
        with a pattern presenter."""



class frequency_mapper(PatternGenerator):
    """Activates a generator sheet at the specified frequency (for all latencies).
    It does so by translating from a frequency to a y position and presenting a line
    at that y position (of the specified size)."""
    
    __abstract = True
            
    size = param.Number(default=0.01, bounds=(0.0,None), softbounds=(0.0,1.0), 
        doc="Thickness (width) of the frequency band for every presentation.")
        
    frequency_spacing = param.Array(default=None,
        doc="""The spacing of the available frequency range, this allows us to define
        (and hence map) a non linear spacing by specifying the frequency value at each
        sheet unit.""")
    
    
    def __init__(self, **params):
        super(frequency_mapper, self).__init__(**params) 
        self.frequency_spacing = round(self.frequency_spacing)

            
    def getFrequency(self):
        sheet_range = self.bounds.lbrt()
        y_range = sheet_range[3] - sheet_range[1]
        
        index = ((self.y - sheet_range[1]) / y_range) * (len(self.frequency_spacing) - 1)

        return self.frequency_spacing[index]
        
        
    def setFrequency(self, new_frequency):
        index = nonzero(self.frequency_spacing >= new_frequency)[0][0]
        
        sheet_range = self.bounds.lbrt()
        y_range = sheet_range[3] - sheet_range[1]
        
        ratio = (len(self.frequency_spacing) - 1) / y_range
        
        y = (index / ratio) + sheet_range[1]
        setattr(self, 'y', y)


    frequency = property(getFrequency, setFrequency, "The frequency at which to present.")
    
    
    def function(self, p):
        return line(self.pattern_y, p.size, 0.001)



class measure_frequency_preference(MeasureResponseCommand):
    """Measure a best frequency preference and selectivity map for auditory neurons."""
        
    display = param.Boolean(True) 
    static_parameters = param.List(default=["scale", "offset"])        
    
    
    def _feature_list(self,p):
        input_sheets = topo.sim.objects(GeneratorSheet).values()
        
        # BK-NOTE: if anyone wants to generalise this method for heterogeneous sheets, by all means do so,
        # for my personal use the additional code required to generalise was overkill.
        for sheet in range(1, len(input_sheets)-1):
            assert input_sheets[sheet].bounds == input_sheets[sheet-1].bounds
            assert input_sheets[sheet].xdensity == input_sheets[sheet-1].xdensity
            assert input_sheets[sheet].ydensity == input_sheets[sheet-1].ydensity
        
        divisions = float(input_sheets[0].ydensity)

        try:
            min_frequency = input_sheets[0].input_generator.min_frequency
            max_frequency = input_sheets[0].input_generator.max_frequency
            frequency_spacing = input_sheets[0].input_generator.frequency_spacing
        except AttributeError:
            min_frequency = 1
            max_frequency = int(divisions) + 1
            frequency_spacing = linspace(min_frequency, max_frequency, num=divisions+1, endpoint=True)
            self.warning("Input generator is missing min_frequency, max_frequency, or frequency_spacing - will present linearly from", str(min_frequency), "to", str(max_frequency), "instead.")

        generator = frequency_mapper(size=1.0/divisions, frequency_spacing=frequency_spacing)
        self.pattern_presenter = PatternPresenter2(pattern_generator=generator)
        
        return [Feature(name="frequency", range=(min_frequency,max_frequency), step=1)]


pg= create_plotgroup(name='Frequency Preference and Selectivity', category="Auditory",
    pre_plot_hooks=[measure_frequency_preference.instance()], normalize='Individually',
    doc='Measure a best frequency preference and selectivity map for auditory neurons.')

pg.add_plot('[Frequency Preference]', [('Strength','FrequencyPreference')])
pg.add_plot('[Frequency Selectivity]', [('Strength','FrequencySelectivity')])



class log_frequency_mapper(PatternGenerator):
    """Activates a generator sheet at the specified frequency (for all latencies).
        It does so by translating from a frequency to a y position and presenting a line
        at that y position (of the specified size)."""

    __abstract = True
    
    size = param.Number(default=0.01, bounds=(0.0,None), softbounds=(0.0,1.0), 
        doc="Thickness (width) of the frequency band for every presentation.")
    
    frequency_spacing = param.Array(default=None,
        doc="""The spacing of the available frequency range, this allows us to define
            (and hence map) a non linear spacing by specifying the frequency value at each
            sheet unit.""")
    
    
    def __init__(self, **params):
        super(log_frequency_mapper, self).__init__(**params) 
        self.frequency_spacing = round(self.frequency_spacing)
    
    
    def getFrequency(self):
        sheet_min = (self.bounds.lbrt())[1]
        index = (self.y - sheet_min) * (self.frequency_spacing.size - 1.0)
        return self.frequency_spacing[index]
    
    
    def setFrequencyBand(self, new_frequency_band):
        sheet_min = (self.bounds.lbrt())[1]
        y = sheet_min + (new_frequency_band / (self.frequency_spacing.size - 1.0))
        setattr(self, 'y', y)
    
    
    frequency = property(getFrequency, setFrequencyBand, "The log frequency band at which to present.")
    
    
    def function(self, p):
        return line(self.pattern_y, p.size, 0.001)
        
        
        
class measure_log_frequency_preference(MeasureResponseCommand):
    """Measure a best frequency preference and selectivity map for auditory neurons."""
    
    display = param.Boolean(True) 
    static_parameters = param.List(default=["scale", "offset"])        
    
    
    def _feature_list(self,p):
        input_sheets = topo.sim.objects(GeneratorSheet).values()
        
        # BK-NOTE: if anyone wants to generalise this method for heterogeneous sheets, by all means do so,
        # for my personal use the additional code required to generalise was overkill.
        for sheet in range(1, len(input_sheets)-1):
            assert input_sheets[sheet].bounds == input_sheets[sheet-1].bounds
            assert input_sheets[sheet].xdensity == input_sheets[sheet-1].xdensity
            assert input_sheets[sheet].ydensity == input_sheets[sheet-1].ydensity
        
        divisions = float(input_sheets[0].ydensity)
        
        try:
            min_frequency = input_sheets[0].input_generator.min_frequency
            max_frequency = input_sheets[0].input_generator.max_frequency
            frequency_spacing = input_sheets[0].input_generator.frequency_spacing
        except AttributeError:
            min_frequency = 1
            max_frequency = int(divisions) + 1
            frequency_spacing = linspace(min_frequency, max_frequency, num=divisions+1, endpoint=True)
            self.warning("Input generator is missing min_frequency, max_frequency, or frequency_spacing - will present linearly from", str(min_frequency), "to", str(max_frequency), "instead.")
        
        generator = log_frequency_mapper(size=1.0/divisions, frequency_spacing=frequency_spacing)
        self.pattern_presenter = PatternPresenter2(pattern_generator=generator)        
        
        return [Feature(name="frequency", range=(0,divisions), step=1)]


pg= create_plotgroup(name='Log Frequency Band Preference and Selectivity', category="Auditory",
                     pre_plot_hooks=[measure_log_frequency_preference.instance()], normalize='Individually',
                     doc='Measure a best frequency preference and selectivity map for auditory neurons (presents frequencies with logarithmically increasing bandwidth).')

pg.add_plot('[Log Frequency Band Preference]', [('Strength','FrequencyPreference')])
pg.add_plot('[Log Frequency Band Selectivity]', [('Strength','FrequencySelectivity')])



class latency_mapper(PatternGenerator):
    """Activates a generator sheet at the specified latency (for all frequencies).
    It does so by translating from a latency to an x position and presenting a line
    at that x position (of the specified size)."""
    
    __abstract = True

    size = param.Number(default=0.01, bounds=(0.0,None), softbounds=(0.0,1.0), 
        doc="Thickness (width) of the latency band for every presentation.")
        
    min_latency = param.Integer(default=0, bounds=(0,None), inclusive_bounds=(True,False),
        doc="""Smallest latency on the generator sheet.""")

    max_latency = param.Integer(default=100, bounds=(0,None), inclusive_bounds=(False,False),
        doc="""Largest latency on the generator sheet.""")


    def getLatency(self):    
        latency_range = self.max_latency - self.min_latency
        
        sheet_range = self.bounds.lbrt()
        x_range = sheet_range[2] - sheet_range[0]
        
        ratio = latency_range / x_range
        
        return ((self.x - sheet_range[0]) * ratio) + self.min_latency
        
        
    def setLatency(self, new_latency):
        latency_range = self.max_latency - self.min_latency
        
        sheet_range = self.bounds.lbrt()
        x_range = sheet_range[2] - sheet_range[0]
        
        ratio = latency_range / x_range
        
        x = ((new_latency-self.min_latency) / ratio) + sheet_range[0]
        setattr(self, 'x', x)


    latency = property(getLatency, setLatency, "The latency at which to present.")           
    
    
    def function(self, p):
        return line(self.pattern_x, p.size, 0.001)



class measure_latency_preference(MeasureResponseCommand):
    """Measure a best onset latency preference and selectivity map for auditory neurons."""
        
    display = param.Boolean(True) 
    static_parameters = param.List(default=["scale", "offset"])        

    def _feature_list(self,p):
        input_sheets = topo.sim.objects(GeneratorSheet).values()
        
        # BK-NOTE: if anyone wants to generalise this method for heterogeneous sheets, by all means do so,
        # for my personal use the additional code required to generalise was overkill.
        for sheet in range(1, len(input_sheets)-1):
            assert input_sheets[sheet].bounds == input_sheets[sheet-1].bounds
            assert input_sheets[sheet].xdensity == input_sheets[sheet-1].xdensity
            assert input_sheets[sheet].ydensity == input_sheets[sheet-1].ydensity
        
        divisions = float(input_sheets[0].xdensity)

        try:
            min_latency = input_sheets[0].input_generator.min_latency
            max_latency = input_sheets[0].input_generator.max_latency
        except AttributeError:
            min_latency = 0
            max_latency = int(divisions)
            self.warning("Input generator is missing min_latency or max_latency, setting to", str(min_latency), "&", str(max_latency),"respectively.")

        generator = latency_mapper(size=1.0/divisions, min_latency=min_latency, max_latency=max_latency)
        self.pattern_presenter = PatternPresenter2(pattern_generator=generator)
        
        return [Feature(name="latency", range=(min_latency,max_latency), step=1)]


pg= create_plotgroup(name='Latency Preference and Selectivity', category="Auditory",
    pre_plot_hooks=[measure_latency_preference.instance()], normalize='Individually',
    doc='Measure a best onset latency preference and selectivity map for auditory neurons.')

pg.add_plot('[Latency Preference]', [('Strength','LatencyPreference')])
pg.add_plot('[Latency Selectivity]', [('Strength','LatencySelectivity')])



import types
__all__ = list(set([k for k,v in locals().items()
                    if isinstance(v,types.FunctionType) or 
                    (isinstance(v,type) and issubclass(v,ParameterizedFunction))
                    and not v.__name__.startswith('_')]))
__all__ += [
    "PatternPresenter2",
    "frequency_mapper",
    "gaussian_corner",
    "latency_mapper",
    "log_frequency_mapper",
]



"""
PlotgroupPanels for displaying ProjectionSheet plotgroups.

$Id$
"""


import ImageTk
### JCALERT! Try not to have to use chain and delete this import.
from itertools import chain
from Tkinter import Canvas
from numpy import sometrue

import param

import topo
from topo.base.cf import CFProjection
from topo.base.projection import ProjectionSheet, Projection
from topo.misc.generatorsheet import GeneratorSheet
        
from templateplotgrouppanel import TemplatePlotGroupPanel



def cmp_projections(p1,p2):
    """
    Comparison function for Plots.
    It compares the precedence number first and then the src_name and name attributes.
    """
    if p1.src.precedence != p2.src.precedence:
        return cmp(p1.src.precedence,p2.src.precedence)
    else:
        return cmp(p1,p2)


UNIT_PADDING = 1



class ProjectionSheetPanel(TemplatePlotGroupPanel):
    """
    Abstract base class for panels relating to ProjectionSheets.
    """
    __abstract = True

    sheet_type = ProjectionSheet
    projection_type = Projection

    @classmethod
    def valid_context(cls):
        """
        Return True if there is at least one instance of
        projection_type among all projections in the simulation.
        """
        for p in chain(*[sheet.projections().values()
                         for sheet in topo.sim.objects(cls.sheet_type).values()]):
            if isinstance(p,cls.projection_type):
                return True

        return False


    def __init__(self,master,plotgroup,**params):
        super(ProjectionSheetPanel,self).__init__(master,plotgroup,**params)

        self.plotgroup.auto_refresh=False
        
        self.pack_param('sheet',parent=self.control_frame_3,
            on_modify=self.sheet_change,side='left',expand=1,
            widget_options={'new_default':True,
                            'sort_fn_args':{'cmp':lambda x, y: cmp(-x.precedence,-y.precedence)}})

        
    def setup_plotgroup(self):
        super(ProjectionSheetPanel,self).setup_plotgroup()
        self.populate_sheet_param()


    def sheet_change(self):
        self.refresh_plots()


    def populate_sheet_param(self):
        sheets = [s for s in topo.sim.objects(self.sheet_type).values()
                  if sometrue([isinstance(p,self.projection_type)
                               for p in s.in_connections])]
        self.plotgroup.params()['sheet'].objects = sheets
        self.plotgroup.sheet = sheets[0] # CB: necessary?



class ProjectionActivityPanel(ProjectionSheetPanel):

    def __init__(self,master,plotgroup,**params):       
        super(ProjectionActivityPanel,self).__init__(master,plotgroup,**params)
        self.auto_refresh = True

    def _plot_title(self):
        return "Activity in projections to %s at time %s"%(self.plotgroup.sheet.name,topo.sim.timestr(self.plotgroup.time))



class UnitsPanel(ProjectionSheetPanel):


    def __init__(self,master,plotgroup,**params):
        self.initial_args=params # CEBALERT: store the initial arguments so we can get sheet,x,y in
                                 # sheet_change if any of them were specified. Isn't there a cleaner
                                 # way?
        super(UnitsPanel,self).__init__(master,plotgroup,**params)
        self.pack_param('x',parent=self.control_frame_4,on_set=self.refresh_plots)
        self.pack_param('y',parent=self.control_frame_4,on_set=self.refresh_plots)
        self.sheet_change()


##############################################################################
        # CEBALERT:
        # - Need to couple taggedslider to a Number parameter in a better way
        # somewhere else.
        # - Clean up or document: passing the params, setting the bounds
        # 
        # Also:        
        # e.g. bound on parameter is 0.5 but means <0.5, taggedslider
        #   still lets you set to 0.5 -> error
            
    def sheet_change(self):
        # CEBHACKALERT: get an inconsequential but scary
        # cf-out-of-range error if you e.g. set y < -0.4 on sheet V1
        # and then change to V2 (which has smaller bounds).
        # x and y don't seem to be updated in time...
        #self.x,self.y = 0.0,0.0

        # CEBALERT: need to crop x,y (for e.g. going to smaller sheet) rather
        # than set to 0
    
        if 'sheet' in self.initial_args: self.sheet=self.initial_args['sheet']

        for coord in ['x','y']:
            self._tkvars[coord].set(self.initial_args.get(coord,0.0))
          
        l,b,r,t = self.sheet.bounds.lbrt()
        # CEBALERT: see "CEBERRORALERT: doesn't take account of
        # exclusive bounds" in topo/param/__init.__.py.
        D=0.0000000001
        bounds = {'x':(l,r-D),
                  'y':(b,t-D)}

        inclusive_bounds = {'x':(True,False),  # GUI knows about exclusive sheet bounds
                            'y':(False,True)}

        for coord in ['x','y']:
            param_obj=self.get_parameter_object(coord)                
            param_obj.bounds = bounds[coord]
            param_obj.inclusive_bounds = inclusive_bounds[coord]
            
            # (method can be called before x,y widgets added)
            if coord in self.representations:
                w=self.representations[coord]['widget']
                w.set_bounds(param_obj.bounds[0],param_obj.bounds[1],
                             inclusive_bounds=param_obj.inclusive_bounds)
                w.tag_set()

        self.initial_args = {} # reset now we've used them
        super(UnitsPanel,self).sheet_change()
##############################################################################

          
class ConnectionFieldsPanel(UnitsPanel):

    projection_type = CFProjection

    def __init__(self,master,plotgroup,**params):
        super(ConnectionFieldsPanel,self).__init__(master,plotgroup,**params)
        self.pack_param('situate',parent=self.control_frame_3,on_set=self.situate_change,side='left',expand=1)
        
    def situate_change(self):
        self.redraw_plots()

    def _plot_title(self):
        return 'Connection Fields of %s unit (%4.3f,%4.3f) at time %s' % \
               (self.sheet.name, self.plotgroup.x,self.plotgroup.y,
                topo.sim.timestr(self.plotgroup.time))


class PlotMatrixPanel(ProjectionSheetPanel):
    """
    PlotGroupPanel for visualizing an array of bitmaps, such as for
    a projection involving a matrix of units.
    """
    
    gui_desired_maximum_plot_height = param.Integer(default=5,bounds=(0,None),doc="""
        Value to provide for PlotGroup.desired_maximum_plot_height for
        PlotGroups opened by the GUI.  Determines the initial, default
        scaling for the PlotGroup.""")

    def sheet_change(self): # don't want to refresh_plots (measure new data) each time
        self.redraw_plots()

    def refresh(self,update=True):
        super(PlotMatrixPanel,self).refresh(update)
        # take the size of the plot as the desired size
        self.plotgroup.update_maximum_plot_height()
        self.desired_maximum_plot_height = self.plotgroup.maximum_plot_height

    
    def display_plots(self):
        """
        CFProjectionPanel requires a 2D grid of plots.
        """
        plots=self.plotgroup.plots
        # Generate the zoomed images.
        self.zoomed_images = [ImageTk.PhotoImage(p.bitmap.image)
                              for p in plots]
        old_canvases = self.canvases

        self.canvases = [Canvas(self.plot_container,
                                width=image.width(),
                                height=image.height(),
                                borderwidth=1,highlightthickness=0,
                                relief='groove')
                         for image in self.zoomed_images]

        # Lay out images
        for i,image,canvas in zip(range(len(self.zoomed_images)),
                                  self.zoomed_images,self.canvases):
            canvas.grid(row=i//self.plotgroup.proj_plotting_shape[1],
                        column=i%self.plotgroup.proj_plotting_shape[1],
                        padx=UNIT_PADDING,pady=UNIT_PADDING)
            canvas.create_image(1,1,anchor='nw',image=image)


        # Delete old ones.  This may resize the grid.
        for c in old_canvases:
            c.grid_forget()

        self._add_canvas_bindings()


    def display_labels(self):
        """Do not display labels for these plots."""
        pass

    
class RFProjectionPanel(PlotMatrixPanel):

    sheet_type = ProjectionSheet

    def __init__(self,master,plotgroup,**params):
        super(RFProjectionPanel,self).__init__(master,plotgroup,**params)
        self.pack_param('input_sheet',parent=self.control_frame_3,
                        on_modify=self.redraw_plots,side='left',expand=1)
                        
        self.pack_param('density',parent=self.control_frame_4)

    def setup_plotgroup(self):
        super(RFProjectionPanel,self).setup_plotgroup()
        self.populate_input_sheet_param()

    def populate_input_sheet_param(self):
        sheets = topo.sim.objects(GeneratorSheet).values()
        self.plotgroup.params()['input_sheet'].objects = sheets
        self.plotgroup.input_sheet=sheets[0]

    def _plot_title(self):
        return 'RFs of %s on %s at time %s'%(self.sheet.name,self.plotgroup.input_sheet.name,
                                             topo.sim.timestr(self.plotgroup.time))
    

class ProjectionPanel(PlotMatrixPanel):
    def __init__(self,master,plotgroup,**params):
        super(ProjectionPanel,self).__init__(master,plotgroup,**params)
        self.pack_param('projection',parent=self.control_frame_3,
                        on_modify=self.redraw_plots,side='left',expand=1,
                        widget_options={'sort_fn_args':{'cmp':cmp_projections},
                                        'new_default':True})
        
        self.pack_param('density',parent=self.control_frame_4) 

        
    def _plot_title(self):
        return self.projection.name + ' projection from ' + self.projection.src.name + ' to ' \
               + self.sheet.name + ' at time ' + topo.sim.timestr(self.plotgroup.time)

    def setup_plotgroup(self):
        super(ProjectionPanel,self).setup_plotgroup()
        self.populate_projection_param()

        
    def sheet_change(self):
        self.refresh_projections()
        super(ProjectionPanel,self).sheet_change()


    def populate_projection_param(self):
        prjns = [proj for proj in self.plotgroup.sheet.projections().values()
                 if isinstance(proj,self.projection_type)]
        self.plotgroup.params()['projection'].objects = prjns
        self.plotgroup.projection = prjns[0] # CB: necessary?


    def refresh_projections(self):
        self.populate_projection_param()

        self.update_selector('projection')

##         #################
##         # CEBALERT: How do you change list of tkinter.optionmenu options? Use pmw's optionmenu?
##         # Or search the web for a way to alter the list in the tkinter one.
##         # Currently, replace widget completely: looks bad and is complex.
##         # When fixing, remove try/except marked by the 'for projectionpanel' CEBALERT in
##         # tkparameterizedobject.py.
##         if 'projection' in self.representations:
##             w  = self.representations['projection']['widget']
##             l  = self.representations['projection']['label']
##             l.destroy()
##             w.destroy()
##             self.pack_param('projection',parent=self.representations['projection']['frame'],
##                             on_modify=self.refresh_plots,side='left',expand=1,
##                             widget_options={'sort_fn_args':{'cmp':cmp_projections},
##                                             'new_default':True})
##         #################


            
class CFProjectionPanel(ProjectionPanel):
    """
    Panel for displaying CFProjections.
    """

    projection_type = CFProjection

    def __init__(self,master,plotgroup,**params):
        super(CFProjectionPanel,self).__init__(master,plotgroup,**params)
        self.pack_param('situate',parent=self.control_frame_3,on_set=self.situate_change,side='left',expand=1)

    def situate_change(self):
        self.redraw_plots()
    



## Need to add test file:
# check projection, sheet ordering
# check sheet changes, check projection changes






"""
Tests for the PlotFileSaver classes. Also tests analysis commands.

$Id$
"""

import unittest
import os
import tempfile
import shutil
import glob

from param import normalize_path

from topo.base.simulation import Simulation
from topo.base.cf import CFSheet, CFProjection
from topo.sheet import GeneratorSheet

from topo.plotting.plotfilesaver import PlotGroupSaver

from topo.plotting.plotgroup import PlotGroup

from topo.command.analysis import save_plotgroup


import __main__
exec "from topo.command.analysis import *" in __main__.__dict__


class TestPlotGroupSaverBase(unittest.TestCase):

    def exists(self,name):
        target = os.path.join(normalize_path.prefix,name)
        files = glob.glob(os.path.join(normalize_path.prefix,"*"))
        self.assert_(os.path.exists(target),
                     "'%s' not among '%s'"%(os.path.basename(target),
                                            [os.path.basename(f) for f in files]))

    def setUp(self):
        self.original_output_path = normalize_path.prefix
        normalize_path.prefix = tempfile.mkdtemp()
        self.sim = Simulation(register=True,name="testplotfilesaver")
        self.sim['A'] = GeneratorSheet(nominal_density=2)
        self.sim['B'] = CFSheet(nominal_density=2)
        self.sim.connect('A','B',connection_type=CFProjection,name='Afferent')        

    def tearDown(self):
        shutil.rmtree(normalize_path.prefix)
        normalize_path.prefix = self.original_output_path
        

class TestPlotGroupSaver(TestPlotGroupSaverBase):

    def test_activity_saving(self):
        save_plotgroup('Activity')
        self.exists("testplotfilesaver_000000.00_A_Activity.png")
        self.exists("testplotfilesaver_000000.00_B_Activity.png")

    def test_orientation_preference_saving(self):
        save_plotgroup('Orientation Preference')
        self.exists("testplotfilesaver_000000.00_B_Orientation_Preference.png")
        self.exists("testplotfilesaver_000000.00_B_Orientation_PreferenceAndSelectivity.png")
        self.exists("testplotfilesaver_000000.00_B_Orientation_Selectivity.png")
        self.exists("testplotfilesaver_000000.00__Color_Key.png")

    def test_cf_saving(self):
        save_plotgroup("Connection Fields",sheet=self.sim['B'])
        self.exists("testplotfilesaver_000000.00_Afferent_(from_A).png")

        
class TestCFProjectionPlotGroupSaver(TestPlotGroupSaverBase):

    def test_cfprojection_saving(self):
        save_plotgroup('Projection',
                       projection=self.sim['B'].projections('Afferent'))
        self.exists("testplotfilesaver_000000.00_B_Afferent.png")


###########################################################

cases = [TestPlotGroupSaver,TestCFProjectionPlotGroupSaver]

suite = unittest.TestSuite()
suite.addTests([unittest.makeSuite(case) for case in cases])

if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(suite)




"""
Unit tests for bitmap.py

Constructs a few bitmap objects but does not display them or compare them
to a base image.  Primary purpose of testing is to verify that the interface
to the PIL imaging library does all the expected operations such as loading
jpg.

Should be extended as more functionality is added to the Bitmap
classes.

$Id$
"""

import topo
from topo.plotting.bitmap import *
from topo.plotting.bitmap import hsv_to_rgb
import Image
import numpy.oldnumeric as Numeric
import unittest
from param import resolve_path

class TestBitmap(unittest.TestCase):


    def setUp(self):
        """
        Uses testbitmap.jpg in the topographica/tests directory
        """
        miata = Image.open(resolve_path('tests/testbitmap.jpg'))
        miata = miata.resize((miata.size[0]/2,miata.size[1]/2))
        self.rIm, self.gIm, self.bIm = miata.split()
        self.rseq = self.rIm.getdata()
        self.gseq = self.gIm.getdata()
        self.bseq = self.bIm.getdata()
        self.rar = Numeric.array(self.rseq)
        self.gar = Numeric.array(self.gseq)
        self.bar = Numeric.array(self.bseq)
        self.ra = Numeric.reshape(self.rar,miata.size) / 255.0
        self.ga = Numeric.reshape(self.gar,miata.size) / 255.0
        self.ba = Numeric.reshape(self.bar,miata.size) / 255.0


    def test_PaletteBitmap(self):
        p = [j and i for i in range(256) for j in (1,0,0)]
        cmap = PaletteBitmap(self.ra,p)
         # cmap.show()


    def test_HSVBitmap(self):
        a = [j for i in range(16) for j in range(16)]
        b = [i for i in range(16) for j in range(16)]
        c = [max(i,j) for i in range(16) for j in range(16)]
        a = Numeric.reshape(a,(16,16)) / 255.0
        b = Numeric.reshape(b,(16,16)) / 255.0
        c = Numeric.reshape(c,(16,16)) / 255.0
        hsv = HSVBitmap(a,b,c)
        # hsv.show()

    def test_hsv_to_rgb(self):
        h,s,v = 0.0,0.0,0.5
        (r,g,b) = hsv_to_rgb(h,s,v)
        self.assertEqual(r,0.5)
        self.assertEqual(g,0.5)
        self.assertEqual(b,0.5)

        h,s,v = 0.0,0.0,0.01
        (r,g,b) = hsv_to_rgb(h,s,v)
        self.assertEqual(r,0.01)
        self.assertEqual(g,0.01)
        self.assertEqual(b,0.01)

 

    def test_RGBBitmap(self):
        rgb = RGBBitmap(self.ra,self.ga,self.ba)
        # rgb.show()


suite = unittest.TestSuite()
suite.addTest(unittest.makeSuite(TestBitmap))



import unittest
import numpy

from topo.base.simulation import Simulation
from topo.base.boundingregion import BoundingBox
from topo.base.cf import CFIter,MaskedCFIter,ResizableCFProjection,CFSheet

class TestCFIter(unittest.TestCase):

    iter_type = CFIter

    def setUp(self):

        self.sim = Simulation()

        self.sim['Dest'] = CFSheet(nominal_density=10,nominal_bounds=BoundingBox(radius=0.5))
        self.sim['Src'] = CFSheet(nominal_density=10,nominal_bounds=BoundingBox(radius=0.5))

        self.sim.connect('Src','Dest',
                         connection_type = ResizableCFProjection,
                         )

    def test_iterate_all(self):
        """
        Test to make sure the iterator hits every CF
        """
        total = 0
        dest = self.sim['Dest']
        proj = dest.projections()['SrcToDest']
        rows,cols = dest.shape
        iterator = self.iter_type(proj)
        for cf,i in iterator():
            total += 1
            self.failUnless(0 <= i < 100, "CFIter generated bogus CF index")
            cfxy = (proj.X_cf.flat[i],proj.Y_cf.flat[i])
            r,c = i/cols,i%cols
            self.failUnlessEqual(cfxy,dest.matrixidx2sheet(r,c))
        self.failUnlessEqual(total,100)


    def test_iterate_some_nil(self):
        """
        Test to make sure iterator skips nil CFs (i.e cf == None)
        """
        dest = self.sim['Dest']
        proj = dest.projections()['SrcToDest']
        total = 0
        proj.flatcfs[24] = None
        for cf,i in self.iter_type(proj)():
            total += 1
            self.failIfEqual(i,24)
        self.failUnlessEqual(total,99)


    def test_iterate_masked(self):
        """
        Test if iterator skips masked CFs
        """
        total = 0
        dest = self.sim['Dest']
        proj = dest.projections()['SrcToDest']
        dest.mask.data = numpy.zeros(dest.activity.shape)
        dest.mask.data.flat[24] = 1
        for cf,i in self.iter_type(proj)():
            total += 1
            self.failUnlessEqual(i,24)
            self.failUnless(cf is proj.flatcfs[24])
        self.failUnlessEqual(total,1)
        


####
cases = [TestCFIter]

suite = unittest.TestSuite()
suite.addTests([unittest.makeSuite(case) for case in cases])

if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(suite)



"""
Topographica plotting subsystem.

Two-dimensional plots are generated by generic classes for visualizing
Sheets in any model.  The plots are often viewed in a GUI, but they
can potentially also be rendered to bitmapped images to save as files,
or to display on other types of interfaces such as web servers.

The usual way that a plot is specified is using a PlotGroupTemplate,
which specifies a group of related plots.  A PlotGroupTemplate is a
list of PlotTemplates.  Each PlotTemplate will normally produce up to
one plot per Sheet in the network, though in special cases it can
produce more.  The definitions for each of these templates can be done
in advance of defining any particular model, and rarely require any
editing for any particular model.

For more information, see the various modules in this package.

$Id$
"""
__all__ = ['bitmap','plot','plotgroup','plotfilesaver']



"""
ConnectionField and associated classes.

This module defines some basic classes of objects used to create
simulations of cortical sheets that take input through connection
fields that project from other cortical sheets (or laterally from
themselves).

ConnectionField: Holds a single connection field within a
CFProjection.

CFProjection: A set of ConnectionFields mapping from a Sheet into a
ProjectionSheet.

CFSheet: A subclass of ProjectionSheet that provides an interface to
the underlying ConnectionFields in any projection of type
CFProjection.

$Id$
"""


from copy import copy

import numpy
from numpy import abs,array,zeros,where
from numpy.oldnumeric import Float,Float32

import param

import patterngenerator
from patterngenerator import PatternGenerator
from functionfamily import TransferFn,IdentityTF
from functionfamily import LearningFn,Hebbian,IdentityLF
from functionfamily import ResponseFn,DotProduct
from functionfamily import CoordinateMapperFn,IdentityMF
from projection import Projection,ProjectionSheet
from sheetcoords import Slice
from sheetview import UnitView
from boundingregion import BoundingBox,BoundingRegionParameter



def simple_vectorize(fn,num_outputs=1,output_type=object,doc=''):
    """
    Wrapper for Numpy.vectorize to make it work properly with different Numpy versions.
    """

    # Numpy.vectorize returns a callable object that applies the given
    # fn to a list or array.  By default, Numpy.vectorize will call
    # the supplied fn an extra time to determine the output types,
    # which is a big problem for any function with side effects.
    # Supplying arguments is supposed to avoid the problem, but as of
    # Numpy 1.6.1 (and apparently since at least 1.1.1) this feature
    # was broken:
    #
    # $ ./topographica -c "def f(x): print x" -c "import numpy" -c "numpy.vectorize(f,otypes=numpy.sctype2char(object)*1)([3,4])"
    # 3
    # 3
    # 4
    #
    # Numpy 1.7.0 seems to fix the problem:
    # $ ./topographica -c "def f(x): print x" -c "import numpy" -c "numpy.vectorize(f,otypes=numpy.sctype2char(object)*1)([3,4])"
    # 3
    # 4
    #
    # To make it work with all versions of Numpy, we use
    # numpy.vectorize as-is for versions > 1.7.0, and a nasty hack for
    # previous versions.

    # Simple Numpy 1.7.0 version:
    if int(numpy.version.version[0]) >= 1 and int(numpy.version.version[2]) >= 7:
        return numpy.vectorize(fn,otypes=numpy.sctype2char(output_type)*num_outputs, doc=doc)

    # Otherwise, we have to mess with Numpy's internal data structures to make it work.
    vfn = numpy.vectorize(fn,doc=doc)
    vfn.nout=num_outputs # number of outputs of fn
    output_typecode = numpy.sctype2char(output_type)
    vfn.otypes=output_typecode*num_outputs # typecodes of outputs of fn
    import inspect

    try:
        fn_code = fn.func_code if hasattr(fn,'func_code') else fn.__call__.func_code
    except:
        raise TypeError("Couldn't find code of %s"%fn)

    fn_args = inspect.getargs(fn_code)[0]
    extra = 1 if fn_args[0]=='self' else 0
    vfn.lastcallargs=len(fn_args)-extra # num args of fn
    return vfn



#: Specified explicitly when creating weights matrix - required
#: for optimized C functions.
weight_type = Float32


class NullCFError(ValueError):
    """
    Error thrown when trying to create an empty CF.
    """
    def __init__(self,x,y,input,rows,cols):
        ValueError.__init__(self,"ConnectionField at (%s,%s) (input_sheet=%s) has a zero-sized weights matrix (%s,%s); you may need to supply a larger bounds_template or increase the density of the sheet."%(x,y,input,rows,cols))
    
                 
class ConnectionField(object):
    """
    A set of weights on one input Sheet.

    Each ConnectionField contributes to the activity of one unit on
    the output sheet, and is normally used as part of a Projection
    including many other ConnectionFields.
    """
    __slots__ = ['weights','input_sheet_slice','mask',
                 '_has_norm_total','_norm_total']

    def __get_norm_total(self):
        """
        Return the stored norm_value, if any, or else the current sum of the weights.
        See the norm_total property for more details.
        """
        # The actual value is cached in _norm_total.
        if self._has_norm_total[0]>0:
            return self._norm_total[0]
        else:
            # CEBALERT: what was I playing with for this before?
            return abs(self.weights).sum()
            
    def __set_norm_total(self,new_norm_total):
        """
        Set an explicit value to be returned by norm_total.
        See the norm_total property for more details.
        """
        self._has_norm_total[0] = 1
        self._norm_total[0] = new_norm_total

    def __del_norm_total(self):
        """
        Delete any cached norm_total that may have been set.
        See the norm_total property for more details.
        """
        self._has_norm_total[0] = 0


    # CB: Accessing norm_total as a property from the C code takes
    # about 2% of run time for 90 iterations of lissom_oo_or. (As of
    # r8139, using floating-point simulation time.)
    norm_total = property(__get_norm_total,__set_norm_total,__del_norm_total,
        """
        The norm_total property returns a value useful in computing
        a sum-based weight normalization.

        By default, the value returned is simply the current sum of
        the connection weights.  However, another value can be
        substituted by setting norm_total explicitly, and this cached
        value will then be returned instead.

        This mechanism has two main purposes.  First, it allows a
        learning function to cache the sum value for an output
        function to use later without computation, which can result in
        significant time savings.  Second, the extra level of
        indirection allows the sum value to be manipulated before it
        is used, to implement operations like joint normalization
        across corresponding CFs in multiple Projections.

        Apart from such cases, norm_total can be ignored.
        
        Note that every person who uses a class that sets or gets
        norm_total must be very careful to ensure that stale values
        will never be accessed.  A good way to do this is to make sure
        that the value is only set just before it will be used, and
        deleted as soon as it has been accessed.
        
        WARNING: Any c-optimized code can bypass this property and
        access directly _has_norm_total, _norm_total
       
        """)


    def get_bounds(self,input_sheet):
        return self.input_sheet_slice.compute_bounds(input_sheet)

    # CEBALERT:
    # template and mask: usually created ONCE by CFProjection and
    # specified as a Slice and array (respectively). Otherwise,
    # can be specified as BoundingBox and patterngenerator.

    # Note that BoundingBox() is ok for a default even though it's
    # mutable because we copy it inside init.  Constant() is ok too
    # because mask and weights_generator are not modified.
    def __init__(self,input_sheet,x=0.0,y=0.0,template=BoundingBox(radius=0.1),
                 weights_generator=patterngenerator.Constant(),
                 mask=patterngenerator.Constant(),
                 output_fns=None,min_matrix_radius=1):
        """
        Create weights at the specified (x,y) location on the
        specified input_sheet.
        
        The supplied template (if a BoundingRegion) is converted to a
        Slice, moved to the specified (x,y) location, and then the
        weights pattern is drawn inside by the weights_generator.

        Note that if the appropriate template Slice is already known,
        then it can be passed in instead of a BoundingRegion template.
        This slice will then be used directly, instead of converting
        the template into a Slice.

        The supplied template object itself will not be modified (it
        is copied before use).
                
        The mask allows the weights to be limited to being non-zero in
        a subset of the rectangular weights area.  The actual mask
        used is a view of the given mask created by cropping to the
        boundaries of the input_sheet, so that the weights all
        correspond to actual locations in the input sheet.  For
        instance, if a circular pattern of weights is desired, the
        mask should have a disk-shaped pattern of elements with value
        1, surrounded by elements with the value 0.  If the CF extends
        over the edge of the input sheet then the weights will
        actually be half-moon (or similar) rather than circular.
        """
        #print "Create CF",input_sheet.name,x,y,"template=",template,"wg=",weights_generator,"m=",mask,"ofs=",output_fns,"min r=",min_matrix_radius

        template = copy(template)

        if not isinstance(template,Slice):
            template = Slice(template,input_sheet,force_odd=True,
                             min_matrix_radius=min_matrix_radius)

        # Note: if passed in, mask is shared between CFs (but not if created here)
        if not hasattr(mask,'view'):
            mask = _create_mask(mask,template.compute_bounds(input_sheet),
        # CEBALERT: it's not really worth adding more ALERTs on this
        # topic, but...there's no way for the CF to control autosize
        # and threshold.
                               input_sheet,True,0.5) 


        # CB: has to be set for C code. Can't be initialized at the
        # class level, or it would become a read-only class attribute
        # (because it's a slot:
        # http://docs.python.org/reference/datamodel.html). Can we
        # somehow avoid having to think about _has_norm_total in the
        # python code? Could the C code initialize this value?
        self._has_norm_total=array([0],dtype=numpy.int32)
        self._norm_total=array([0.0],dtype=float)

        if output_fns is None:
            output_fns = []
            
        # CEBALERT: now even more confusing; weights_slice is
        # different from input_sheet_slice. At least need to rename.
        weights_slice = self._create_input_sheet_slice(input_sheet,x,y,template,min_matrix_radius)

        # CBNOTE: this would be clearer (but not perfect, and probably slower)
        # m = mask_template[self.weights_slice()]
        self.mask = weights_slice.submatrix(mask)  # view of original mask
        self.mask = array(self.mask,copy=1) # CEBALERT: why is this necessary?

        # (without it, optimized learning function creates artifacts in CFs at
        # left and right edges of sheet, at some densities)
        
        # CBENHANCEMENT: might want to do something about a size
        # that's specified (right now the size is assumed to be that
        # of the bounds)
        # shouldn't be extra computation of boundingbox because it's gone from Slice.__init__; could avoid extra lookups by getting straight from slice
        w = weights_generator(x=x,y=y,bounds=self.get_bounds(input_sheet),
                              xdensity=input_sheet.xdensity,
                              ydensity=input_sheet.ydensity,
                              mask=self.mask)

        # CEBALERT: unnecessary copy! Pass type to PG & have it draw
        # in that.  (Should be simple, except making it work for all
        # the PG subclasses that override array creation in various
        # ways (producing or using inconsistent types) turned out to
        # be too painful.)
        self.weights = w.astype(weight_type)

        # CEBHACKALERT: the system of masking through multiplication
        # by 0 works for now, while the output_fns are all
        # multiplicative.  But in the long run we need a better way to
        # apply the mask.  The same applies anywhere the mask is used,
        # including in learningfn/. We should investigate masked
        # arrays (from numpy).
        for of in output_fns:
            of(self.weights)        


    # CB: can this be renamed to something better?
    def _create_input_sheet_slice(self,input_sheet,x,y,template,min_matrix_radius):
        """
        Create the input_sheet_slice, which provides the appropriate
        Slice for this CF on the input_sheet (as well as providing
        this CF's exact bounds).

        Also creates the weights_slice, which provides the Slice for
        this weights matrix (in case it must be cropped at an edge).
        """
        # copy required because the template gets modified here but
        # needs to be used again
        input_sheet_slice = copy(template)
        input_sheet_slice.positionedcrop(x,y,input_sheet)
        input_sheet_slice.crop_to_sheet(input_sheet)

        # weights matrix cannot have a zero-sized dimension (could
        # happen at this stage because of cropping)
        nrows,ncols = input_sheet_slice.shape_on_sheet()
        if nrows<1 or ncols<1:
            raise NullCFError(x,y,input_sheet,nrows,ncols)

        self.input_sheet_slice = input_sheet_slice

        # not copied because we don't use again
        template.positionlesscrop(x,y,input_sheet)
        return template


    # CEBALERT: unnecessary method; can use something like
    # activity[cf.input_sheet_slice()]
    def get_input_matrix(self, activity):
        # CBNOTE: again, this might be clearer (but probably slower):
        # activity[self.input_sheet_slice()]
        return self.input_sheet_slice.submatrix(activity)



class CFPResponseFn(param.Parameterized):
    """
    Map an input activity matrix into an output matrix using the CFs
    in a CFProjection.

    Objects in this hierarchy of callable function objects compute a
    response matrix when given an input pattern and a set of
    ConnectionField objects.  Typically used as part of the activation
    function for a neuron, computing activation for one Projection.

    Objects in this class must support being called as a function with
    the arguments specified below, and are assumed to modify the
    activity matrix in place.
    """
    __abstract=True

    def __call__(self, iterator, input_activity, activity, strength, **params):
        raise NotImplementedError


class CFPRF_Plugin(CFPResponseFn):
    """
    Generic large-scale response function based on a simple single-CF function.

    Applies the single_cf_fn to each CF in turn.  For the default
    single_cf_fn of DotProduct(), does a basic dot product of each CF with the
    corresponding slice of the input array.  This function is likely
    to be slow to run, but it is easy to extend with any arbitrary
    single-CF response function.

    The single_cf_fn must be a function f(X,W) that takes two
    identically shaped matrices X (the input) and W (the
    ConnectionField weights) and computes a scalar activation value
    based on those weights.
    """
    single_cf_fn = param.ClassSelector(ResponseFn,default=DotProduct(),
        doc="Accepts a ResponseFn that will be applied to each CF individually.")
    
    def __call__(self, iterator, input_activity, activity, strength):
        single_cf_fn = self.single_cf_fn
        for cf,i in iterator():
            X = cf.input_sheet_slice.submatrix(input_activity)
            activity.flat[i] = single_cf_fn(X,cf.weights)
        activity *= strength


class CFPLearningFn(param.Parameterized):
    """
    Compute new CFs for a CFProjection based on input and output activity values.

    Objects in this hierarchy of callable function objects compute a
    new set of CFs when given input and output patterns and a set of
    ConnectionField objects.  Used for updating the weights of one
    CFProjection.

    Objects in this class must support being called as a function with
    the arguments specified below.
    """
    __abstract = True
        

    def constant_sum_connection_rate(self,n_units,learning_rate):
        """ 
        Return the learning rate for a single connection assuming that
        the total rate is to be divided evenly among all the units in
        the connection field.
        """
        return float(learning_rate)/n_units


    # JABALERT: Should the learning_rate be a parameter of this object instead of an argument?
    def __call__(self, iterator, input_activity, output_activity, learning_rate, **params):
        """
        Apply this learning function to the given set of ConnectionFields,
        and input and output activities, using the given learning_rate.
        """
        raise NotImplementedError


class CFPLF_Identity(CFPLearningFn):
    """CFLearningFunction performing no learning."""
    single_cf_fn = param.ClassSelector(LearningFn,default=IdentityLF(),constant=True)
  
    def __call__(self, iterator, input_activity, output_activity, learning_rate, **params):
        pass


class CFPLF_Plugin(CFPLearningFn):
    """CFPLearningFunction applying the specified single_cf_fn to each CF."""
    single_cf_fn = param.ClassSelector(LearningFn,default=Hebbian(),
        doc="Accepts a LearningFn that will be applied to each CF individually.")
    def __call__(self, iterator, input_activity, output_activity, learning_rate, **params):
        """Apply the specified single_cf_fn to every CF."""
        single_connection_learning_rate = self.constant_sum_connection_rate(iterator.proj_n_units,learning_rate)
        # avoid evaluating these references each time in the loop
        single_cf_fn = self.single_cf_fn

        for cf,i in iterator():
            single_cf_fn(cf.get_input_matrix(input_activity),
                         output_activity.flat[i], cf.weights, 
                         single_connection_learning_rate)
            cf.weights *= cf.mask                


class CFPOutputFn(param.Parameterized):
    """
    Type for an object that applies some operation (typically something
    like normalization) to all CFs in a CFProjection for which the specified
    mask (typically the activity at the destination of this projection)
    is nonzero.
    """
    __abstract = True

    def __call__(self, iterator, **params):
        """Operate on each CF for which the mask is nonzero."""
        raise NotImplementedError


class CFPOF_Plugin(CFPOutputFn):
    """
    Applies the specified single_cf_fn to each CF in the CFProjection
    for which the mask is nonzero.
    """
    single_cf_fn = param.ClassSelector(TransferFn,default=IdentityTF(),
        doc="Accepts a TransferFn that will be applied to each CF individually.")
    
    def __call__(self, iterator, **params):
        if type(self.single_cf_fn) is not IdentityTF:
            single_cf_fn = self.single_cf_fn

            for cf,i in iterator():
                single_cf_fn(cf.weights)
                del cf.norm_total


class CFPOF_Identity(CFPOutputFn):
    """
    CFPOutputFn that leaves the CFs unchanged.

    Must never be changed or subclassed, because it might never
    be called. (I.e., it could simply be tested for and skipped.)
    """
    single_cf_fn = param.ClassSelector(TransferFn,default=IdentityTF(),constant=True)
    
    def __call__(self, iterator, **params):
        pass


# CB: need to make usage of 'src' and 'input_sheet' consistent between
# ConnectionField and CFProjection (i.e. pick one of them).
class CFProjection(Projection):
    """
    A projection composed of ConnectionFields from a Sheet into a ProjectionSheet.

    CFProjection computes its activity using a response_fn of type
    CFPResponseFn (typically a CF-aware version of mdot) and output_fns 
    (typically none).  The initial contents of the 
    ConnectionFields mapping from the input Sheet into the target
    ProjectionSheet are controlled by the weights_generator, cf_shape,
    and weights_output_fn parameters, while the location of the
    ConnectionField is controlled by the coord_mapper parameter.

    Any subclass has to implement the interface
    activate(self,input_activity) that computes the response from the
    input and stores it in the activity array.
    """

    response_fn = param.ClassSelector(CFPResponseFn,
        default=CFPRF_Plugin(),
        doc='Function for computing the Projection response to an input pattern.')
    
    cf_type = param.Parameter(default=ConnectionField,constant=True,
        doc="Type of ConnectionField to use when creating individual CFs.")

    # JPHACKALERT: Not all support for null CFs has been implemented.
    # CF plotting and C-optimized CFPxF_ functions need
    # to be fixed to support null CFs without crashing.    
    allow_null_cfs = param.Boolean(default=False,
        doc="Whether or not the projection can have entirely empty CFs")

    nominal_bounds_template = BoundingRegionParameter(
        default=BoundingBox(radius=0.1),doc="""
        Bounds defining the Sheet area covered by a prototypical ConnectionField.
        The true bounds will differ depending on the density (see create_slice_template()).""")
    
    weights_generator = param.ClassSelector(PatternGenerator,
        default=patterngenerator.Constant(),constant=True,
        doc="Generate initial weights values.")

    cf_shape = param.ClassSelector(PatternGenerator,
        default=patterngenerator.Constant(),constant=True,
        doc="Mask pattern to define the shape of the connection fields.")

    same_cf_shape_for_all_cfs = param.Boolean(default=True,doc="""
        Whether or not to share a single cf_shape mask for all CFs.
        If True, the cf_shape is evaluated only once and shared for
        all CFs, which saves computation time and memory.  If False,
        the cf_shape is evaluated once for each CF, allowing each to
        have its own shape.""")
    
    learning_fn = param.ClassSelector(CFPLearningFn,
        default=CFPLF_Plugin(),
        doc='Function for computing changes to the weights based on one activation step.')

    # JABALERT: Shouldn't learning_rate be owned by the learning_fn?
    learning_rate = param.Number(default=0.0,softbounds=(0,100),doc="""
        Amount of learning at each step for this projection, specified
        in units that are independent of the density of each Sheet.""")

    weights_output_fns = param.HookList(default=[CFPOF_Plugin()],
        class_=CFPOutputFn,
        doc='Functions applied to each CF after learning.')

    strength = param.Number(default=1.0,doc="""
        Global multiplicative scaling applied to the Activity of this Sheet.""")

    coord_mapper = param.ClassSelector(CoordinateMapperFn,
        default=IdentityMF(),
        doc='Function to map a projected coordinate into the target sheet.')

    # CEBALERT: this is temporary (allows c++ matching in certain
    # cases).  We will allow the user to override the mask size, but
    # by offering a scaling parameter.
    autosize_mask = param.Boolean(
        default=True,constant=True,precedence=-1,doc="""
        Topographica sets the mask size so that it is the same as the connection field's
        size, unless this parameter is False - in which case the user-specified size of
        the cf_shape is used. In normal usage of Topographica, this parameter should
        remain True.""")

    mask_threshold = param.Number(default=0.5,constant=True,doc="""
        If a unit is above this value in the cf_shape mask, it is
        included; otherwise it is excluded from the mask.""")

    apply_output_fns_init=param.Boolean(default=True,doc="""
        Whether to apply the output function to connection fields (e.g. for 
        normalization) when the CFs are first created.""")

    min_matrix_radius = param.Integer(default=1,bounds=(0,None),doc="""
        Enforced minimum for radius of weights matrix.
        The default of 1 gives a minimum matrix of 3x3. 0 would
        allow a 1x1 matrix.""")


    precedence = param.Number(default=0.8)


    def __init__(self,initialize_cfs=True,**params):
        """
        Initialize the Projection with a set of cf_type objects
        (typically ConnectionFields), each located at the location
        in the source sheet corresponding to the unit in the target
        sheet. The cf_type objects are stored in the 'cfs' array.

        The nominal_bounds_template specified may be altered: the
        bounds must be fitted to the Sheet's matrix, and the weights
        matrix must have odd dimensions. These altered bounds are
        passed to the individual connection fields.

        A mask for the weights matrix is constructed. The shape is
        specified by cf_shape; the size defaults to the size
        of the nominal_bounds_template.
        """
        super(CFProjection,self).__init__(**params)

        self.weights_generator.set_dynamic_time_fn(None,sublistattr='generators')
        # get the actual bounds_template by adjusting a copy of the
        # nominal_bounds_template to ensure an odd slice, and to be
        # cropped to sheet if necessary
        self._slice_template = Slice(copy(self.nominal_bounds_template),
                                     self.src,force_odd=True,
                                     min_matrix_radius=self.min_matrix_radius)

        self.bounds_template = self._slice_template.compute_bounds(self.src)
        
        self.mask_template = _create_mask(self.cf_shape,self.bounds_template,
                                         self.src,self.autosize_mask,
                                         self.mask_threshold)

        self.n_units = self._calc_n_units()
        
        if initialize_cfs:
            self._create_cfs()

            
        ### JCALERT! We might want to change the default value of the
        ### input value to self.src.activity; but it fails, raising a
        ### type error. It probably has to be clarified why this is
        ### happening
        self.input_buffer = None
        self.activity = array(self.dest.activity)


    def _generate_coords(self):
        X,Y = self.dest.sheetcoords_of_idx_grid()
        vectorized_coord_mapper = simple_vectorize(self.coord_mapper,
                                                   num_outputs=2,
                                                   # CB: could switch to float32?
                                                   output_type=float) 
        return vectorized_coord_mapper(X,Y)


    # CB: should be _initialize_cfs() since we already have 'initialize_cfs' flag?
    def _create_cfs(self):
        vectorized_create_cf = simple_vectorize(self._create_cf)
        #:
        self.cfs = vectorized_create_cf(*self._generate_coords())
        self.flatcfs = list(self.cfs.flat)

        
    def _create_cf(self,x,y):
        """
        Create a ConnectionField at x,y in the src sheet.
        """
        # (to restore would need to have an r,c counter)
        # self.debug("Creating CF(%d,%d) from src (%.3f,%.3f) to  dest (%.3f,%.3f)"%(r,c,x_cf,y_cf,x,y))

        try:
            if self.apply_output_fns_init:
                ofs = [wof.single_cf_fn for wof in self.weights_output_fns]
            else:
                ofs = []

            if self.same_cf_shape_for_all_cfs:
                mask_template = self.mask_template
            else:
                mask_template = _create_mask(self.cf_shape,self.bounds_template,
                                            self.src,self.autosize_mask,
                                            self.mask_threshold)

            CF = self.cf_type(self.src,x=x,y=y,
                              template=self._slice_template,
                              weights_generator=self.weights_generator,
                              mask=mask_template, 
                              output_fns=ofs,
                              min_matrix_radius=self.min_matrix_radius)
        except NullCFError:
            if self.allow_null_cfs:
                CF = None
            else:
                raise
        
        return CF



    def _calc_n_units(self):
        """Return the number of unmasked units in a typical ConnectionField."""      

        return min(len(self.mask_template.ravel().nonzero()[0]),
                   # CEBALERT: if the mask_template is bigger than the
                   # src sheet (e.g.  conn radius bigger than src
                   # radius), return the size of the source sheet
                   self.src.shape[0]*self.src.shape[1])


    def cf(self,r,c):
        """Return the specified ConnectionField"""
        # CB: should we offer convenience cf(x,y) (i.e. sheetcoords) method instead?
        self.warning("CFProjection.cf(r,c) is deprecated: use cfs[r,c] instead")
        return self.cfs[r,c]


    def cf_bounds(self,r,c):
        """Return the bounds of the specified ConnectionField."""
        return self.cfs[r,c].get_bounds(self.src)


    def get_view(self, sheet_x, sheet_y, timestamp):
        """
        Return a single connection field UnitView, for the unit
        located nearest to sheet coordinate (sheet_x,sheet_y).
        """
        matrix_data = zeros(self.src.activity.shape,Float) 
        (r,c) = self.dest.sheet2matrixidx(sheet_x,sheet_y)
        r1,r2,c1,c2 = self.cfs[r,c].input_sheet_slice
        matrix_data[r1:r2,c1:c2] = self.cfs[r,c].weights

        # CB: the following would be equivalent with Slice __call__
        
        # cf = self.cf(self.dest.sheet2matrixidx(sheet_x,sheet_y))
        # matrix_data = numpy.zeros(self.src.activity.shape,Numeric.Float)
        # matrix_data[cf.input_sheet_slice()]=cf.weights

        return UnitView((matrix_data,self.src.bounds),sheet_x,sheet_y,self,timestamp)


    def activate(self,input_activity):
        """Activate using the specified response_fn and output_fn."""
        self.input_buffer = input_activity
        self.activity *=0.0
        self.response_fn(MaskedCFIter(self), input_activity, self.activity, self.strength)
        for of in self.output_fns:
            of(self.activity)


    # CEBALERT: should add active_units_mask to match
    # apply_learn_output_fns.  
    def learn(self):
        """
        For a CFProjection, learn consists of calling the learning_fn.
        """
        # Learning is performed if the input_buffer has already been set,
        # i.e. there is an input to the Projection.
        if self.input_buffer != None:
            self.learning_fn(MaskedCFIter(self),self.input_buffer,self.dest.activity,self.learning_rate)
       

    # CEBALERT: called 'learn' output fns here, but called 'weights' output fns
    # elsewhere (mostly). Change all to 'learn'?
    def apply_learn_output_fns(self,active_units_mask=True):
        """
        Apply the weights_output_fns to each unit.

        If active_units_mask is True, inactive units will be skipped.
        """
        for of in self.weights_output_fns:
            of(MaskedCFIter(self,active_units_mask=active_units_mask))


    # CEBALERT: see gc alert in simulation.__new__
    def _cleanup(self):
        for cf in self.cfs.flat:
            # cf could be None or maybe something else
            if hasattr(cf,'input_sheet'):
                cf.input_sheet=None
            if hasattr(cf,'input_sheet_slice'):
                cf.input_sheet_slice=None
            if hasattr(cf,'weights_slice'):
                cf.weights_slice=None


    def n_bytes(self):
        # Could also count the input_sheet_slice
        rows,cols=self.cfs.shape
        return super(CFProjection,self).n_bytes() + \
               sum([cf.weights.nbytes + 
                    cf.mask.nbytes
                    for cf,i in CFIter(self,ignore_sheet_mask=True)()])


    def n_conns(self):
        # Counts non-masked values, if mask is available; otherwise counts 
        # weights as connections if nonzero
        rows,cols=self.cfs.shape
        return sum([len((cf.mask if cf.mask is not None else cf.weights).ravel().nonzero()[0])
                    for cf,i in MaskedCFIter(self)()])


# CEB: have not yet decided proper location for this method
# JAB: should it be in PatternGenerator?
def _create_mask(shape,bounds_template,sheet,autosize=True,threshold=0.5):
    """
    Create the mask (see ConnectionField.__init__()).
    """
    # Calculate the size & aspect_ratio of the mask if appropriate;
    # mask size set to be that of the weights matrix
    if hasattr(shape, 'size') and autosize:
        l,b,r,t = bounds_template.lbrt()
        shape.size = t-b
        shape.aspect_ratio = (r-l)/shape.size

    # Center mask to matrixidx center
    center_r,center_c = sheet.sheet2matrixidx(0,0)
    center_x,center_y = sheet.matrixidx2sheet(center_r,center_c)

    mask = shape(x=center_x,y=center_y,
                 bounds=bounds_template,
                 xdensity=sheet.xdensity,
                 ydensity=sheet.ydensity)

    mask = where(mask>=threshold,mask,0.0)

    # CB: unnecessary copy (same as for weights)
    return mask.astype(weight_type)



import numpy
class CFIter(object):
    """
    Iterator to walk through all ConnectionFields of all neurons in
    the destination Sheet of the given CFProjection.  Each iteration
    yields the tuple (cf,i) where cf is the ConnectionField at
    position i in the projection's flatcfs list. 

    If active_units_mask is True, inactive units will be skipped. If
    ignore_sheet_mask is True, even units excluded by the sheet mask
    will be included.
    """
    
    # CB: as noted elsewhere, rename active_units_mask (to e.g.
    # ignore_inactive_units).
    def __init__(self,cfprojection,active_units_mask=False,ignore_sheet_mask=False):

        self.flatcfs = cfprojection.flatcfs
        self.activity = cfprojection.dest.activity
        self.mask = cfprojection.dest.mask
        self.cf_type = cfprojection.cf_type
        self.proj_n_units = cfprojection.n_units
        self.allow_skip_non_responding_units = cfprojection.dest.allow_skip_non_responding_units

        self.active_units_mask = active_units_mask
        self.ignore_sheet_mask = ignore_sheet_mask

    def __nomask(self):
        # return an array indicating all units should be processed 
        
        # dtype for C functions. 
        # could just be flat.
        return numpy.ones(self.activity.shape,dtype=self.activity.dtype)

    # CEBALERT: make _
    def get_sheet_mask(self):
        if not self.ignore_sheet_mask:
            return self.mask.data
        else: 
            return self.__nomask()

    # CEBALERT: make _ (and probably drop '_mask').
    def get_active_units_mask(self):
        if self.allow_skip_non_responding_units and self.active_units_mask:
            return self.activity
        else:
            return self.__nomask()

    # CEBALERT: rename?
    def get_overall_mask(self):
        """
        Return an array indicating whether or not each unit should be
        processed.
        """
        # JPHACKALERT: Should really check for the existence of the
        # mask, rather than checking its type. This is a hack to
        # support higher-order projections whose dest is a CF, instead
        # of a sheet.  The right thing to do is refactor so that CF
        # masks and  SheetMasks are subclasses of an abstract Mask
        # type so that they support the same interfaces.
        #
        # CEBALERT: put back when supporting neighborhood masking
        # (though preferably do what Jeff suggests instead)
#        if isinstance(self.proj.dest.mask,SheetMask):
#            return get_active_units_mask()
#        else:

        # CB: note that it's faster for our optimized C functions to
        # combine the masks themselves, rather than using this method.
        sheet_mask = self.get_sheet_mask()
        active_units_mask = self.get_active_units_mask()
        return numpy.logical_and(sheet_mask,active_units_mask)
    

    def __call__(self):
        mask = self.get_overall_mask()
        for i,cf in enumerate(self.flatcfs):
            if cf is not None:
                if mask.flat[i]:
                    yield cf,i

# CEBALERT: remove this once MaskedCFIter has been replaced elsewhere.
MaskedCFIter = CFIter


### We don't really need this class; its methods could probably be
### moved up to ProjectionSheet, because they may in fact be valid for
### all ProjectionSheets. But we're leaving it here, because it is
### likely to be useful in the future.
class CFSheet(ProjectionSheet):
    """
    A ProjectionSheet providing access to the ConnectionFields in its CFProjections.

    CFSheet is a Sheet built from units indexed by Sheet coordinates
    (x,y).  Each unit can have one or more ConnectionFields on another
    Sheet (via this sheet's CFProjections).  Thus CFSheet is a more
    concrete version of a ProjectionSheet; a ProjectionSheet does not
    require that there be units or weights of any kind.  Unless you
    need access to the underlying ConnectionFields for visualization
    or analysis, CFSheet and ProjectionSheet are interchangeable.
    """

    measure_maps = param.Boolean(True,doc="""
        Whether to include this Sheet when measuring various maps to create SheetViews.""")

    precedence = param.Number(0.5)


    def update_unit_view(self,x,y,proj_name=''):
        """
        Creates the list of UnitView objects for a particular unit in this CFSheet.
        (There is one UnitView for each Projection to this CFSheet).

        Each UnitView is then added to the sheet_views of its source sheet.
        It returns the list of all UnitViews for the given unit.
        """     
        for p in self.in_connections:
            if not isinstance(p,CFProjection):
                self.debug("Skipping non-CFProjection "+p.name)
            elif proj_name == '' or p.name==proj_name:
                v = p.get_view(x,y,self.simulation.time())
                src = v.projection.src
                key = ('Weights',v.projection.dest.name,v.projection.name,x,y)
                v.proj_src_name = v.projection.src.name
                src.sheet_views[key] = v


    ### JCALERT! This should probably be deleted...
    def release_unit_view(self,x,y):
        self.release_sheet_view(('Weights',x,y))




class ResizableCFProjection(CFProjection):
    """
    A CFProjection with resizable weights.
    """
    # Less efficient memory usage than CFProjection because it stores
    # the (x,y) position of each ConnectionField.

    
    def _generate_coords(self):
        # same as super's, but also stores the coords.

        # CB: this is storing redundant info because generate_coords()
        # returns output from mgrid. Might be better to store the 1d x
        # and y coords, and generate the grids when needed?
        self.X_cf,self.Y_cf = super(ResizableCFProjection,self)._generate_coords()
        return self.X_cf,self.Y_cf
        

    ### This could be changed into a special __set__ method for
    ### bounds_template, instead of being a separate function, but
    ### having it be explicit like this might be clearer.
    ###
    ### This implementation is fairly slow, and for some algorithms
    ### that rely on changing the bounds frequently, it may be worth
    ### re-implementing it in C.
    def change_bounds(self, nominal_bounds_template):
        """
        Change the bounding box for all of the ConnectionFields in this Projection.

        Calls change_bounds() on each ConnectionField.

        Currently only allows reducing the size, but should be
        extended to allow increasing as well.
        """
        slice_template = Slice(copy(nominal_bounds_template),
                               self.src,force_odd=True,
                               min_matrix_radius=self.min_matrix_radius)

        bounds_template = slice_template.compute_bounds(self.src)

        if not self.bounds_template.containsbb_exclusive(bounds_template):
            if self.bounds_template.containsbb_inclusive(bounds_template):
                self.debug('Initial and final bounds are the same.')
            else:
                self.warning('Unable to change_bounds; currently allows reducing only.')
            return

        # it's ok so we can store the bounds and resize the weights
        mask_template = _create_mask(self.cf_shape,bounds_template,self.src,
                                    self.autosize_mask,self.mask_threshold)

        self.mask_template = mask_template
        self.n_units = self._calc_n_units()
        self.nominal_bounds_template = nominal_bounds_template

        self.bounds_template = bounds_template
        self._slice_template = slice_template

        cfs = self.cfs
        rows,cols = cfs.shape
        output_fns = [wof.single_cf_fn for wof in self.weights_output_fns]

        for r in xrange(rows):
            for c in xrange(cols):
                xcf,ycf = self.X_cf[0,c],self.Y_cf[r,0]
                # CB: listhack - loop is candidate for replacement by numpy fn
                self._change_cf_bounds(cfs[r,c],input_sheet=self.src,
                                       x=xcf,y=ycf,
                                       template=slice_template,
                                       mask=mask_template,
                                       output_fns=output_fns,
                                       min_matrix_radius=self.min_matrix_radius)


    def change_density(self, new_wt_density):
        """
        Rescales the weight matrix in place, interpolating or resampling as needed.
        
        Not yet implemented.
        """
        raise NotImplementedError


    def _change_cf_bounds(self,cf,input_sheet,x,y,template,mask,output_fns=None,min_matrix_radius=1):
        """
        Change the bounding box for this ConnectionField.

        Discards weights or adds new (zero) weights as necessary,
        preserving existing values where possible.

        Currently only supports reducing the size, not increasing, but
        should be extended to support increasing as well.

        Note that the supplied template will be modified, so if you're
        also using them elsewhere you should pass copies.
        """
        if output_fns is None:
            output_fns = []
            
        # CEBALERT: re-write to allow arbitrary resizing
        or1,or2,oc1,oc2 = cf.input_sheet_slice

        weights_slice = cf._create_input_sheet_slice(input_sheet,x,y,copy(template),min_matrix_radius)
                    
        r1,r2,c1,c2 = cf.input_sheet_slice


        if not (r1 == or1 and r2 == or2 and c1 == oc1 and c2 == oc2):
            # CB: note that it's faster to copy (i.e. replacing copy=1 with copy=0
            # below slows down change_bounds().
            cf.weights = array(cf.weights[r1-or1:r2-or1,c1-oc1:c2-oc1],copy=1)
            # (so the obvious choice,
            # cf.weights=cf.weights[r1-or1:r2-or1,c1-oc1:c2-oc1],
            # is also slower).

            cf.mask = weights_slice.submatrix(mask)
            cf.mask = array(cf.mask,copy=1) # CB: why's this necessary?
                                                # (see ALERT in __init__)


            cf.weights *= cf.mask
            for of in output_fns:
                of(cf.weights)
            del cf.norm_total



"""
A family of function objects for transforming a matrix generated from some other function.

Output functions are useful for neuron activation functions,
normalization of matrices, etc.  They come in two varieties:
OutputFunction, and CFPOutputFunction.  An OutputFunction
(e.g. PiecewiseLinear) applies to one matrix of any type, such as an
activity matrix or a set of weights.  To apply an OutputFunction to
all of the weight matrices in an entire CFProjection, an
OutputFunction can be plugged in to CFPOF_Plugin.  CFPOF_Plugin is one
example of a CFPOutputFunction, which is a function that works with
the entire Projection at once.

Any new output functions added to this directory will automatically
become available for any model.

$Id$
"""

import copy

import param

import numpy, numpy.random
from numpy import exp,zeros,ones,power

import topo
import topo.base.functionfamily
from topo.base.sheet import activity_type
from topo.base.arrayutil import clip_lower

# Imported here so that all TransferFns will be in the same package
from imagen.transferfn import TransferFn,IdentityTF,Threshold  # pyflakes:ignore (API import)
from imagen.transferfn import BinaryThreshold,DivisiveNormalizeL1  # pyflakes:ignore (API import)
from imagen.transferfn import DivisiveNormalizeL2,DivisiveNormalizeLinf # pyflakes:ignore (API import)
from imagen.transferfn import DivisiveNormalizeLp # pyflakes:ignore (API import)

# CEBHACKALERT: these need to respect the mask - which will be passed in.


class PiecewiseLinear(TransferFn):
    """ 
    Piecewise-linear TransferFn with lower and upper thresholds.
    
    Values below the lower_threshold are set to zero, those above
    the upper threshold are set to 1.0, and those in between are
    scaled linearly.
    """
    lower_bound = param.Number(default=0.0,softbounds=(0.0,1.0))
    upper_bound = param.Number(default=1.0,softbounds=(0.0,1.0))
    
    def __call__(self,x):
        fact = 1.0/(self.upper_bound-self.lower_bound)        
        x -= self.lower_bound
        x *= fact
        x.clip(0.0,1.0,out=x)



class Sigmoid(TransferFn):
    """ 
    Sigmoidal (logistic) transfer function: 1/(1+exp-(r*x+k)).

    As defined in Jochen Triesch, ICANN 2005, LNCS 3696 pp. 65-70. 
    The parameters control the growth rate (r) and the x position (k)
    of the exponential.
    
    This function is a special case of the GeneralizedLogistic
    function, with parameters r=r, l=0, u=1, m=-k/2r, and b=1.  See
    Richards, F.J. (1959), A flexible growth function for empirical
    use. J. Experimental Botany 10: 290--300, 1959.
    http://en.wikipedia.org/wiki/Generalised_logistic_curve
    """
    
    r = param.Number(default=1,doc="Parameter controlling the growth rate")
    k = param.Number(default=0,doc="Parameter controlling the x-postion")
    
    def __call__(self,x):
        x_orig = copy.copy(x)
        x *= 0.0
        x += 1.0 / (1.0 + exp(-(self.r*x_orig+self.k)))


                  
class NakaRushton(TransferFn):
    #JABALERT: Please write the equation into words in the docstring, as in Sigmoid.
    """
    Naka-Rushton curve.

    From Naka, K. and Rushton, W. (1996), S-potentials from luminosity
    units in the retina of fish (Cyprinidae). J. Physiology 185:587-599.

    The Naka-Rushton curve has been shown to be a good approximation
    of constrast gain control in cortical neurons.  The input of the
    curve is usually contrast, but under the assumption that the
    firing rate of a model neuron is directly proportional to the
    contrast, it can be used as a TransferFn for a Sheet.
    
    The parameter c50 corresponds to the contrast at which the half of
    the maximal output is reached.  For a Sheet TransferFn this translates
    to the input for which a neuron will respond with activity 0.5.
    """
    
    c50 = param.Number(default=0.1, doc="""
        The input of the neuron at which it responds at half of its
        maximal firing rate (1.0).""")

    e = param.Number(default=1.0,doc="""
        The exponent of the input x.""")

    #JABALERT: (pow(x_orig,self.e) should presumably be done only once, using a temporary
    def __call__(self,x):
        #print 'A:', x
        #print 'B:', pow(x,self.e) / (pow(x,self.e) + pow(self.c50,self.e))
        x_orig = copy.copy(x)
        x *= 0
        x += pow(x_orig,self.e) / (pow(x_orig,self.e) + pow(self.c50,self.e))



class GeneralizedLogistic(TransferFn):
    """ 
    The generalized logistic curve (Richards' curve): y = l + (u /(1 + b * exp(-r*(x-2*m))^(1/b))).

    The logistic curve is a flexible function for specifying a
    nonlinear growth curve using five parameters:

    * l: the lower asymptote
    * u: the upper asymptote minus l
    * m: the time of maximum growth
    * r: the growth rate
    * b: affects near which asymptote maximum growth occurs

    From Richards, F.J. (1959), A flexible growth function for empirical
    use. J. Experimental Botany 10: 290--300.
    http://en.wikipedia.org/wiki/Generalised_logistic_curve
    """
    
    # JABALERT: Reword these to say what they are, not what they
    # control, if they are anything that can be expressed naturally.
    # E.g. is l a parameter controlling the lower asymptote, or is it
    # simply the lower asymptote?  If it's the lower asymptote, say
    # doc="Lower asymptote.".  Only if the parameter's relationship to
    # what it controls is very indirect should it be worded as below.
    l = param.Number(default=1,doc="Parameter controlling the lower asymptote.")
    u = param.Number(default=1,doc="Parameter controlling the upper asymptote (upper asymptote minus lower asymptote.")
    m = param.Number(default=1,doc="Parameter controlling the time of maximum growth.")
    r = param.Number(default=1,doc="Parameter controlling the growth rate.")
    b = param.Number(default=1,doc="Parameter which affects near which asymptote maximum growth occurs.")
    
    def __call__(self,x):
        x_orig = copy.copy(x)
        x *= 0.0
        x += self.l + ( self.u /(1 + self.b*exp(-self.r *(x_orig - 2*self.m))**(1 / self.b)) )    



class HalfRectifyAndSquare(TransferFn):
    """
    Transfer function that applies a half-wave rectification (clips at zero)
    and then squares the values.
    """
    t = param.Number(default=0.0,doc="""
        The threshold at which output becomes non-zero.""")
    
    def __call__(self,x):
        x -= self.t
        clip_lower(x,0)
        x *= x



class HalfRectifyAndPower(TransferFn):
    """
    Transfer function that applies a half-wave rectification (i.e.,
    clips at zero), and then raises the result to the e-th power
    (where the exponent e can be selected arbitrarily).
    """
    e = param.Number(default=2.0,doc="""
        The exponent to which the thresholded value is raised.""")
    
    t = param.Number(default=0.0,doc="""
        The threshold level subtracted from x.""")
    
    def __call__(self,x):
        x -= self.t
        clip_lower(x,0)
        a = power(x,self.e)
        x*=0
        x+=a



class ExpLinear(TransferFn):
    """
    Transfer function that is exponential until t from which point it is linear.
    """

    e = param.Number(default=1.0,doc="""
        The exponent of the exponetial part of the curve""")
    t1 = param.Number(default=0.5,doc="""
        The threshold level where function becomes non-zero""")
    t2 = param.Number(default=1.0,doc="""
        The threshold level at which curve becomes linear""")
        
    a = param.Number(default=1.0,doc="""
        The overall scaling of the function""")
    
    def __call__(self,x):
        x-=self.t1
        clip_lower(x,0)
        z = (x>=self.t2)*(1.0*exp(self.e*(-self.t2))+(x-self.t2)) + \
            (x<self.t2)*(exp(self.e*(x-self.t2))-exp(self.e*(-self.t2)))
        x*=0
        x+=self.a*z



class Square(TransferFn):
    """Transfer function that applies a squaring nonlinearity."""

    def __call__(self,x):
        x *= x     
        


# JAALERT: rename to something like PlasticTransferFn
class TransferFnWithState(TransferFn):
    """
    Abstract base class for TransferFns that need to maintain a self.plastic parameter.

    These TransferFns typically maintain some form of internal history
    or other state from previous calls, which can be disabled by
    override_plasticity_state().
    """

    plastic = param.Boolean(default=True, doc="""
        Whether or not to update the internal state on each call.
        Allows plasticity to be turned off during analysis, and then re-enabled.""")

    __abstract = True
    
    def __init__(self,**params):
        super(TransferFnWithState,self).__init__(**params)
        self._plasticity_setting_stack = []


    def override_plasticity_state(self, new_plasticity_state):
        """
        Temporarily disable plasticity of internal state.

        This function should be implemented by all subclasses so that
        after a call, the output should always be the same for any
        given input pattern (apart from true randomness or other
        differences that do not depend on an internal state), and no
        call should have any effect that persists after a subsequent
        restore_plasticity_state() call.
        
        By default, simply saves a copy of the 'plastic' parameter to
        an internal stack (so that it can be restored by
        restore_plasticity_state()), and then sets the plastic
        parameter to the given value (True or False).
        """
        self._plasticity_setting_stack.append(self.plastic)
        self.plastic=new_plasticity_state


    def restore_plasticity_state(self):
        """
        Re-enable plasticity of internal state after an override_plasticity_state call.

        This function should be implemented by all subclasses to
        remove the effect of the most recent override_plasticity_state call,
        i.e. to reenable changes to the internal state, without any
        lasting effect from the time during which plasticity was disabled.

        By default, simply restores the last saved value of the
        'plastic' parameter.
        """
        self.plastic = self._plasticity_setting_stack.pop()                        

    def state_push(self):
        """
        Save the current state onto a stack, to be restored using state_pop.
        
        Subclasses must implement state_push and state_pop if they
        store any lasting state across invocations, so that the result
        of state_pop will be the state that was present at the
        previous state_push.
        """
        pass
    
    def state_pop(self):
        """
        Restore the state saved by the most recent state_push call.
        """
        pass



# CB: it's not ideal that all TransferFnWithRandomState fns have
# the plastic stuff (from TransferFnWithState).
class TransferFnWithRandomState(TransferFnWithState):
    """
    Abstract base class for TransferFns that use a random number generator.
    """

    random_generator = param.Parameter(
        default=numpy.random.RandomState(seed=(10,10)),doc=
        """
        numpy's RandomState provides methods for generating random
        numbers (see RandomState's help for more information).

        Note that all instances of subclasses of
        TransferFnWithRandomState will share this RandomState object,
        and hence its state. To create an instance of an
        TransferFnWithRandomState subclass that has its own state, set
        this parameter on the instance to a new RandomState instance.
        """)
        
    __abstract = True
    
    def __init__(self,**params):
        super(TransferFnWithRandomState,self).__init__(**params)
        self.__random_generators_stack = []

    def state_push(self):
        """
        Save the current random number generator (onto the stack),
        replacing it with a copy.
        """
        self.__random_generators_stack.append(self.random_generator)
        self.random_generator=copy.copy(self.random_generator)
        super(TransferFnWithRandomState,self).state_push()

    def state_pop(self):
        """
        Retrieve the previous random number generator from the stack.
        """
        self.random_generator = self.__random_generators_stack.pop()
        super(TransferFnWithRandomState,self).state_push()
        


class PoissonSample(TransferFnWithRandomState):
    """
    Simulate Poisson-distributed activity with specified mean values.
    
    This transfer function interprets each matrix value as the
    (potentially scaled) rate of a Poisson process and replaces it
    with a sample from the appropriate Poisson distribution.

    To allow the matrix to contain values in a suitable range (such as
    [0.0,1.0]), the input matrix is scaled by the parameter in_scale,
    and the baseline_rate is added before sampling.  After sampling,
    the output value is then scaled by out_scale.  The function thus
    performs this transformation::

      x <- P(in_scale * x + baseline_rate) * out_scale

    where x is a matrix value and P(r) samples from a Poisson
    distribution with rate r.
    """
    
    in_scale = param.Number(default=1.0,doc="""
       Amount by which to scale the input.""")
    
    baseline_rate = param.Number(default=0.0,doc="""
       Constant to add to the input after scaling, resulting in a baseline
       Poisson process rate.""")
    
    out_scale = param.Number(default=1.0,doc="""
       Amount by which to scale the output (e.g. 1.0/in_scale).""")

    def __call__(self,x):
        
        x *= self.in_scale
        x += self.baseline_rate
        sample = self.random_generator.poisson(x,x.shape)
        x *= 0.0
        x += sample
        x *= self.out_scale



# CEBALERT: I think this TF is a bit misleading because it does not
# alter the input array. Is a TransferFn the right thing to use to
# record an average activity? Could consider replacing this with my
# "attribute tracking" code (some of which is not yet in SVN).
class ActivityAveragingTF(TransferFnWithState):
    """
    Calculates the average of the input activity.

    The average is calculated as an exponential moving average, where
    the weighting for each older data point decreases exponentially.
    The degree of weighing for the previous values is expressed as a
    constant smoothing factor.

    The plastic parameter allows the updating of the average values
    to be disabled temporarily, e.g. while presenting test patterns.
    """

    step = param.Number(default=1, doc="""
        How often to update the average.
        For instance, step=1 means to update it every time this TF is
        called; step=2 means to update it every other time.""")
    
    smoothing = param.Number(default=0.9997, doc="""
        The degree of weighting for the previous average, when
        calculating the new average.""")
   
    initial_average=param.Number(default=0, doc="""
        Starting value for the average activity.""")

    
    def __init__(self,**params):
        super(ActivityAveragingTF,self).__init__(**params)
        self.n_step = 0
        self.x_avg=None


    def __call__(self,x):
        if self.x_avg is None:
            self.x_avg=self.initial_average*ones(x.shape, activity_type)         

        # Collect values on each appropriate step
        
        self.n_step += 1
        if self.n_step == self.step:
            self.n_step = 0
            if self.plastic:
                self.x_avg = (1.0-self.smoothing)*x + self.smoothing*self.x_avg



class HomeostaticMaxEnt(TransferFnWithRandomState):
    """
    Implementation of homeostatic intrinsic plasticity from Jochen Triesch,
    ICANN 2005, LNCS 3696 pp.65-70.
    
    A sigmoid activation function is adapted automatically to achieve
    desired average firing rate and approximately exponential
    distribution of firing rates (for the maximum possible entropy).
    
    Note that this TransferFn has state, so the history of calls to it
    will affect future behavior.  The plastic parameter can be used
    to disable changes to the state.
    
    Also calculates average activity as useful debugging information,
    for use with AttributeTrackingTF.  Average activity is calculated as
    an exponential moving average with a smoothing factor (smoothing).
    For more information see:
    NIST/SEMATECH e-Handbook of Statistical Methods, Single Exponential Smoothing
    http://www.itl.nist.gov/div898/handbook/pmc/section4/pmc431.htm
    """

    eta = param.Number(default=0.0002,doc="""
        Learning rate for homeostatic plasticity.""")
    
    mu = param.Number(default=0.01,doc="""
        Target average firing rate.""")
    
    smoothing = param.Number(default=0.9997, doc="""
        Weighting of previous activity vs. current activity when
        calculating the average.""")

    a_init = param.Parameter(default=None,doc="""
        Multiplicative parameter controlling the exponential.""")
   
    b_init = param.Parameter(default=None,doc="""
        Additive parameter controlling the exponential.""")

    step = param.Number(default=1, doc="""
        How often to update the a and b parameters.  
        For instance, step=1 means to update it every time this TF is
        called; step=2 means to update it every other time.""")

    def __init__(self,**params):
        super(HomeostaticMaxEnt,self).__init__(**params)
        self.first_call = True
        self.n_step=0
        self.__current_state_stack=[]
        self.a=None
        self.b=None
        self.y_avg=None

    def __call__(self,x):      
        if self.first_call:
            self.first_call = False
            if self.a_init==None:
                self.a = self.random_generator.uniform(low=10, high=20,size=x.shape)
            else:
                self.a = ones(x.shape, x.dtype.char) * self.a_init
            if self.b_init==None:
                self.b = self.random_generator.uniform(low=-8.0, high=-4.0,size=x.shape)
            else:
                self.b = ones(x.shape, x.dtype.char) * self.b_init
            self.y_avg = zeros(x.shape, x.dtype.char) 

        # Apply sigmoid function to x, resulting in what Triesch calls y
        x_orig = copy.copy(x)
       
        x *= 0.0
        x += 1.0 / (1.0 + exp(-(self.a*x_orig + self.b)))
                

        self.n_step += 1
        if self.n_step == self.step:
            self.n_step = 0
            if self.plastic:                
                self.y_avg = (1.0-self.smoothing)*x + self.smoothing*self.y_avg #Calculate average for use in debugging only
                
                # Update a and b
                self.a += self.eta * (1.0/self.a + x_orig - (2.0 + 1.0/self.mu)*x_orig*x + x_orig*x*x/self.mu)
                self.b += self.eta * (1.0 - (2.0 + 1.0/self.mu)*x + x*x/self.mu)


    def state_push(self):
        self.__current_state_stack.append((copy.copy(self.a), copy.copy(self.b), copy.copy(self.y_avg), copy.copy(self.first_call)))
        super(HomeostaticMaxEnt,self).state_push()

        
    def state_pop(self):
        self.a, self.b, self.y_avg, self.first_call =  self.__current_state_stack.pop()
        super(HomeostaticMaxEnt,self).state_pop()
        


class ScalingTF(TransferFnWithState):
    """
    Scales input activity based on the current average activity (x_avg).

    The scaling is calculated to bring x_avg for each unit closer to a
    specified target average.  Calculates a scaling factor that is
    greater than 1 if x_avg is less than the target and less than 1 if
    x_avg is greater than the target, and multiplies the input
    activity by this scaling factor.

    The plastic parameter allows the updating of the average values
    to be disabled temporarily, e.g. while presenting test patterns.
    """
    
    target = param.Number(default=0.01, doc="""
        Target average activity for each unit.""")

    step=param.Number(default=1, doc="""
        How often to calculate the average activity and scaling factor.""")

    smoothing = param.Number(default=0.9997, doc="""
        Determines the degree of weighting of previous activity vs.
        current activity when calculating the average.""")

    
    def __init__(self,**params):
        super(ScalingTF,self).__init__(**params)
        self.n_step = 0
        self.x_avg=None
        self.sf=None

    def __call__(self,x):
    
        if self.x_avg is None:
            self.x_avg=self.target*ones(x.shape, activity_type)         
        if self.sf is None:
            self.sf=ones(x.shape, activity_type)

        # Collect values on each appropriate step
       
        self.n_step += 1
        if self.n_step == self.step:
            self.n_step = 0
            if self.plastic:
                self.sf *= 0.0
                self.sf += self.target/self.x_avg
                self.x_avg = (1.0-self.smoothing)*x + self.smoothing*self.x_avg
        
        x *= self.sf
              


class Hysteresis(TransferFnWithState):
    """
    Smoothly interpolates a matrix between simulation time steps, with
    exponential falloff.
    """

    time_constant  = param.Number(default=0.3,doc="""
        Controls the time scale of the interpolation.""")
    
    def __init__(self,**params):
        super(Hysteresis,self).__init__(**params)
        self.first_call = True
        self.__current_state_stack=[]
        self.old_a = 0 
        topo.base.functionfamily.PatternDrivenAnalysis.pre_presentation_hooks.append(self.reset)
        
    def __call__(self,x):
        if self.first_call is True:
           self.old_a = x.copy() * 0.0
           self.first_call = False
           
        #if (float(topo.sim.time()) %1.0) <= 0.15: self.old_a =  self.old_a* 0 
        new_a = x.copy()
        self.old_a = self.old_a + (new_a - self.old_a)*self.time_constant
        x*=0
        x += self.old_a
        
    def reset(self):
        self.old_a *= 0

    def state_push(self):
        self.__current_state_stack.append((copy.copy(self.old_a), copy.copy(self.first_call)))
        super(Hysteresis,self).state_push()

    def state_pop(self):
        self.old_a,self.first_call =  self.__current_state_stack.pop()
        super(Hysteresis,self).state_pop()


_public = list(set([_k for _k,_v in locals().items() if isinstance(_v,type) and issubclass(_v,TransferFn)]))


# Automatically discover all .py files in this directory.
import os,fnmatch
__all__ = _public + [f.split('.py')[0] for f in os.listdir(__path__[0]) if fnmatch.fnmatch(f,'[!._]*.py')]
del f,os,fnmatch
